"""This module contains all the code responsible for the HTOP-like interface."""

# Why using asciimatics?
#
# - curses is hard, and not working well on Windows
# - blessings (curses-based) is easier, but does not provide input methods and is not maintained
# - blessed (blessings fork) provides input methods, but they are blocking
# - urwid seems less easy to use than asciimatics, and older
# - clint is not maintained and does not provide input methods
# - prompt_toolkit is designed to build interactive (like, very interactive) command line applications
# - curtsies, pygcurse, unicurses, npyscreen: all based on curses anyway, which does not work well on Windows
#
# Well, asciimatics also provides a "top" example, so...

from __future__ import annotations

import contextlib
import os
import sys
import time
from collections import defaultdict
from pathlib import Path
from typing import TYPE_CHECKING, Callable, ClassVar, Sequence, TypedDict

import pyperclip
import requests
from asciimatics.event import KeyboardEvent, MouseEvent
from asciimatics.screen import ManagedScreen, Screen
from loguru import logger

from aria2p.api import API
from aria2p.utils import get_version, load_configuration

if TYPE_CHECKING:
    from aria2p.downloads import Download


configs = load_configuration()


def key_bind_parser(action: str) -> list[Key]:
    """Return a list of Key instances.

    Parameters:
        action: The action name.

    Returns:
        A list of keys.
    """
    default_bindings = configs["DEFAULT"]["key_bindings"]
    bindings = configs.get("USER", {}).get("key_bindings", default_bindings)

    key_binds = bindings.get(action, default_bindings[action])

    if isinstance(key_binds, list):
        return [Key(k) for k in key_binds]
    return [Key(key_binds)]


def color_palette_parser(palette: str) -> tuple[int, int, int]:
    """Return a color tuple (foreground color, mode, background color).

    Parameters:
        palette: The palette name.

    Returns:
        Foreground color, mode, background color.
    """
    default_colors = configs["DEFAULT"]["colors"]
    colors = configs.get("USER", {}).get("colors", default_colors)

    # get values of colors and modes for ascimatics.screen module
    color_map = {
        "BLACK": Screen.COLOUR_BLACK,
        "WHITE": Screen.COLOUR_WHITE,
        "RED": Screen.COLOUR_RED,
        "CYAN": Screen.COLOUR_CYAN,
        "YELLOW": Screen.COLOUR_YELLOW,
        "BLUE": Screen.COLOUR_BLUE,
        "GREEN": Screen.COLOUR_GREEN,
        "DEFAULT": Screen.COLOUR_DEFAULT,
    }
    mode_map = {
        "NORMAL": Screen.A_NORMAL,
        "BOLD": Screen.A_BOLD,
        "UNDERLINE": Screen.A_UNDERLINE,
        "REVERSE": Screen.A_REVERSE,
    }

    palette_colors = colors.get(palette, default_colors[palette])
    palette_fg, palette_mode, palette_bg = palette_colors.split(" ")

    return (
        color_map[palette_fg],
        mode_map[palette_mode],
        color_map[palette_bg],
    )


class Key:
    """A class to represent an input key."""

    OTHER_KEY_VALUES: ClassVar[dict[str, int]] = {
        "F1": Screen.KEY_F1,
        "F2": Screen.KEY_F2,
        "F3": Screen.KEY_F3,
        "F4": Screen.KEY_F4,
        "F5": Screen.KEY_F5,
        "F6": Screen.KEY_F6,
        "F7": Screen.KEY_F7,
        "F8": Screen.KEY_F8,
        "F9": Screen.KEY_F9,
        "F10": Screen.KEY_F10,
        "F11": Screen.KEY_F11,
        "F12": Screen.KEY_F12,
        "ESC": Screen.KEY_ESCAPE,
        "DEL": Screen.KEY_DELETE,
        "PAGE_UP": Screen.KEY_PAGE_UP,
        "PAGE_DOWN": Screen.KEY_PAGE_DOWN,
        "HOME": Screen.KEY_HOME,
        "END": Screen.KEY_END,
        "LEFT": Screen.KEY_LEFT,
        "UP": Screen.KEY_UP,
        "RIGHT": Screen.KEY_RIGHT,
        "DOWN": Screen.KEY_DOWN,
        "BACK": Screen.KEY_BACK,
        "TAB": Screen.KEY_TAB,
        "SPACE": ord(" "),
        "ENTER": ord("\n"),
    }

    def __init__(self, name: str, value: int | None = None) -> None:
        """Initialize the object.

        Parameters:
            name: The key name.
            value: The key value.
        """
        self.name = name
        if value is None:
            value = self.get_value(name)
        self.value = value

    def get_value(self, name: str) -> int:  # noqa: D102
        try:
            value = ord(name)
        except TypeError:
            value = self.OTHER_KEY_VALUES[name.upper()]

        return value

    def __eq__(self, value: object) -> bool:
        return self.value == value


class Keys:
    """The actions and their shortcuts keys."""

    AUTOCLEAR = key_bind_parser("AUTOCLEAR")
    CANCEL = key_bind_parser("CANCEL")
    ENTER = key_bind_parser("ENTER")
    FILTER = key_bind_parser("FILTER")
    FOLLOW_ROW = key_bind_parser("FOLLOW_ROW")
    HELP = key_bind_parser("HELP")
    MOVE_DOWN = key_bind_parser("MOVE_DOWN")
    MOVE_LEFT = key_bind_parser("MOVE_LEFT")
    MOVE_RIGHT = key_bind_parser("MOVE_RIGHT")
    MOVE_UP = key_bind_parser("MOVE_UP")
    NEXT_SORT = key_bind_parser("NEXT_SORT")
    PREVIOUS_SORT = key_bind_parser("PREVIOUS_SORT")
    PRIORITY_DOWN = key_bind_parser("PRIORITY_DOWN")
    PRIORITY_UP = key_bind_parser("PRIORITY_UP")
    QUIT = key_bind_parser("QUIT")
    REMOVE_ASK = key_bind_parser("REMOVE_ASK")
    REVERSE_SORT = key_bind_parser("REVERSE_SORT")
    SEARCH = key_bind_parser("SEARCH")
    SELECT_SORT = key_bind_parser("SELECT_SORT")
    SETUP = key_bind_parser("SETUP")
    TOGGLE_EXPAND_COLLAPSE_ALL = key_bind_parser("TOGGLE_EXPAND_COLLAPSE_ALL")
    TOGGLE_EXPAND_COLLAPSE = key_bind_parser("TOGGLE_EXPAND_COLLAPSE")
    TOGGLE_RESUME_PAUSE = key_bind_parser("TOGGLE_RESUME_PAUSE")
    TOGGLE_RESUME_PAUSE_ALL = key_bind_parser("TOGGLE_RESUME_PAUSE_ALL")
    TOGGLE_SELECT = key_bind_parser("TOGGLE_SELECT")
    UN_SELECT_ALL = key_bind_parser("UN_SELECT_ALL")
    MOVE_HOME = key_bind_parser("MOVE_HOME")
    MOVE_END = key_bind_parser("MOVE_END")
    MOVE_UP_STEP = key_bind_parser("MOVE_UP_STEP")
    MOVE_DOWN_STEP = key_bind_parser("MOVE_DOWN_STEP")
    RETRY = key_bind_parser("RETRY")
    RETRY_ALL = key_bind_parser("RETRY_ALL")
    ADD_DOWNLOADS = key_bind_parser("ADD_DOWNLOADS")

    @staticmethod
    def names(keys_list: list[Key]) -> list[str]:  # noqa: D102
        return [key.name for key in keys_list]

    @staticmethod
    def values(keys_list: list[Key]) -> list[int]:  # noqa: D102
        return [key.value for key in keys_list]


class Exit(Exception):  # noqa: N818
    """A simple exception to exit the interactive interface."""


class Column:
    """A class to specify a column in the interface.

    It's composed of a header (the string to display on top), a padding (how to align the text),
    and three callable functions to get the text from a Python object, to sort between these objects,
    and to get a color palette based on the text.
    """

    def __init__(
        self,
        header: str,
        padding: str,
        get_text: Callable,
        get_sort: Callable,
        get_palette: Callable,
    ) -> None:
        """Initialize the object.

        Parameters:
            header: The string to display on top.
            padding: How to align the text.
            get_text: Function accepting a Download as argument and returning the text to display.
            get_sort: Function accepting a Download as argument and returning the attribute used to sort.
            get_palette: Function accepting text as argument and returning a palette or a palette identifier.
        """
        self.header = header
        self.padding = padding
        self.get_text = get_text
        self.get_sort = get_sort
        self.get_palette = get_palette


class HorizontalScroll:
    """A wrapper around asciimatics' Screen.print_at and Screen.paint methods.

    It allows scroll the rows horizontally, used when moving left and right:
    the first N characters will not be printed.
    """

    def __init__(self, screen: Screen, scroll: int = 0) -> None:
        """Initialize the object.

        Parameters:
            screen (Screen): The asciimatics screen object.
            scroll (int): Base scroll to use when printing. Will decrease by one with each character skipped.
        """
        self.screen = screen
        self.scroll = scroll

    def set_scroll(self, scroll: int) -> None:
        """Set the scroll value."""
        self.scroll = scroll

    def print_at(self, text: str, x: int, y: int, palette: list | tuple) -> int:
        """Wrapper print_at method.

        Parameters:
            text: Text to print.
            x: X axis position / column.
            y: Y axis position / row.
            palette: A length-3 tuple or a list of length-3 tuples representing asciimatics palettes.

        Returns:
            The number of characters actually printed.
        """
        if self.scroll == 0:
            if isinstance(palette, list):
                self.screen.paint(text, x, y, colour_map=palette)
            else:
                self.screen.print_at(text, x, y, *palette)
            written = len(text)
        else:
            text_length = len(text)
            if text_length > self.scroll:
                new_text = text[self.scroll :]
                written = len(new_text)
                if isinstance(palette, list):
                    new_palette = palette[self.scroll :]
                    self.screen.paint(new_text, x, y, colour_map=new_palette)
                else:
                    self.screen.print_at(new_text, x, y, *palette)
                self.scroll = 0
            else:
                self.scroll -= text_length
                written = 0
        return written


class Palette:
    """A simple class to hold palettes getters."""

    @staticmethod
    def status(value: str) -> str:
        """Return the palette for a STATUS cell."""
        return "status_" + value

    @staticmethod
    def name(value: str) -> str | list[tuple[int, int, int]]:
        """Return the palette for a NAME cell."""
        if value.startswith("[METADATA]"):
            return (
                [(Screen.COLOUR_GREEN, Screen.A_UNDERLINE, Screen.COLOUR_BLACK)] * 10
                + [Interface.palettes["metadata"]] * (len(value.strip()) - 10)
                + [Interface.palettes["row"]]
            )
        return "name"


class Interface:
    """The main class responsible for drawing the HTOP-like interface.

    It should be instantiated with an API instance, and then ran with its `run` method.

    If you want to re-use this class' code to create an HTOP-like interface for another purpose,
    simply change these few things:

    - columns, columns_order and palettes attributes
    - sort and reverse attributes default values
    - get_data method. It should return a list of objects that can be compared by equality (==, __eq__, __hash__)
    - __init__ method to accept other arguments
    - remove/change the few events with "download" or "self.api" in the process_event method
    """

    class State:  # noqa: D106
        MAIN = 0
        HELP = 1
        SETUP = 2
        REMOVE_ASK = 3
        SELECT_SORT = 4
        ADD_DOWNLOADS = 9

    state = State.MAIN
    sleep = 0.005
    frames = 200  # 200 * 0.005 seconds == 1 second
    frame = 0
    focused = 0
    side_focused = 0
    sort = 2
    reverse = True
    x_scroll = 0
    x_offset = 0
    y_offset = 0
    row_offset = 0
    refresh = False
    width: int
    height: int
    screen: Screen
    data: list[Download]
    rows: list[Sequence[str]]
    scroller: HorizontalScroll
    follow = None
    bounds: list[Sequence[int]]

    palettes: ClassVar[dict[str, tuple[int, int, int]]] = defaultdict(lambda: color_palette_parser("UI"))
    palettes.update(
        {
            "ui": color_palette_parser("UI"),
            "header": color_palette_parser("HEADER"),
            "focused_header": color_palette_parser("FOCUSED_HEADER"),
            "focused_row": color_palette_parser("FOCUSED_ROW"),
            "status_active": color_palette_parser("STATUS_ACTIVE"),
            "status_paused": color_palette_parser("STATUS_PAUSED"),
            "status_waiting": color_palette_parser("STATUS_WAITING"),
            "status_error": color_palette_parser("STATUS_ERROR"),
            "status_complete": color_palette_parser("STATUS_COMPLETE"),
            "metadata": color_palette_parser("METADATA"),
            "side_column_header": color_palette_parser("SIDE_COLUMN_HEADER"),
            "side_column_row": color_palette_parser("SIDE_COLUMN_ROW"),
            "side_column_focused_row": color_palette_parser("SIDE_COLUMN_FOCUSED_ROW"),
            "bright_help": color_palette_parser("BRIGHT_HELP"),
        },
    )

    columns_order: ClassVar[list[str]] = ["gid", "status", "progress", "size", "down_speed", "up_speed", "eta", "name"]
    columns: ClassVar[dict[str, Column]] = {
        "gid": Column(
            header="GID",
            padding=">16",
            get_text=lambda d: d.gid,
            get_sort=lambda d: d.gid,
            get_palette=lambda d: "gid",
        ),
        "status": Column(
            header="STATUS",
            padding="<9",
            get_text=lambda d: d.status,
            get_sort=lambda d: d.status,
            get_palette=Palette.status,
        ),
        "progress": Column(
            header="PROGRESS",
            padding=">8",
            get_text=lambda d: d.progress_string(),
            get_sort=lambda d: d.progress,
            get_palette=lambda s: "progress",
        ),
        "size": Column(
            header="SIZE",
            padding=">11",
            get_text=lambda d: d.total_length_string(),
            get_sort=lambda d: d.total_length,
            get_palette=lambda s: "size",
        ),
        "down_speed": Column(
            header="DOWN_SPEED",
            padding=">13",
            get_text=lambda d: d.download_speed_string(),
            get_sort=lambda d: d.download_speed,
            get_palette=lambda s: "down_speed",
        ),
        "up_speed": Column(
            header="UP_SPEED",
            padding=">13",
            get_text=lambda d: d.upload_speed_string(),
            get_sort=lambda d: d.upload_speed,
            get_palette=lambda s: "up_speed",
        ),
        "eta": Column(
            header="ETA",
            padding=">8",
            get_text=lambda d: d.eta_string(precision=2),
            get_sort=lambda d: d.eta,
            get_palette=lambda s: "eta",
        ),
        "name": Column(
            header="NAME",
            padding="100%",
            get_text=lambda d: d.name,
            get_sort=lambda d: d.name,
            get_palette=Palette.name,
        ),
    }

    remove_ask_header = "Remove:"
    remove_ask_rows: ClassVar[list[tuple]] = [
        ("Remove", lambda d: d.remove(force=False, files=False)),
        ("Remove with files", lambda d: d.remove(force=False, files=True)),
        ("Force remove", lambda d: d.remove(force=True, files=False)),
        ("Force remove with files", lambda d: d.remove(force=True, files=True)),
    ]
    last_remove_choice = None

    select_sort_header = "Select sort:"
    select_sort_rows = columns_order

    downloads_uris: list[str]
    downloads_uris_header = (
        f"Add Download: [ Hit ENTER to download; Hit { ','.join(Keys.names(Keys.ADD_DOWNLOADS)) } to download all ]"
    )

    class StateConf(TypedDict):  # noqa: D106
        process_keyboard_event: Callable
        process_mouse_event: Callable
        print_functions: list[Callable]

    def __init__(self, api: API | None = None) -> None:
        """Initialize the object.

        Parameters:
            api: An instance of API.
        """
        if api is None:
            api = API()
        self.api = api

        self.rows = []
        self.data = []
        self.bounds = []
        self.downloads_uris = []
        self.height = 20
        self.width = 80

        # reduce curses' 1 second delay when hitting escape to 25 ms
        os.environ.setdefault("ESCDELAY", "25")

        self.state_mapping: dict[int, Interface.StateConf] = {
            self.State.MAIN: {
                "process_keyboard_event": self.process_keyboard_event_main,
                "process_mouse_event": self.process_mouse_event_main,
                "print_functions": [self.print_table],
            },
            self.State.HELP: {
                "process_keyboard_event": self.process_keyboard_event_help,
                "process_mouse_event": self.process_mouse_event_help,
                "print_functions": [self.print_help],
            },
            self.State.SETUP: {
                "process_keyboard_event": self.process_keyboard_event_setup,
                "process_mouse_event": self.process_mouse_event_setup,
                "print_functions": [],
            },
            self.State.REMOVE_ASK: {
                "process_keyboard_event": self.process_keyboard_event_remove_ask,
                "process_mouse_event": self.process_mouse_event_remove_ask,
                "print_functions": [self.print_remove_ask_column, self.print_table],
            },
            self.State.SELECT_SORT: {
                "process_keyboard_event": self.process_keyboard_event_select_sort,
                "process_mouse_event": self.process_mouse_event_select_sort,
                "print_functions": [self.print_select_sort_column, self.print_table],
            },
            self.State.ADD_DOWNLOADS: {
                "process_keyboard_event": self.process_keyboard_event_add_downloads,
                "process_mouse_event": self.process_mouse_event_add_downloads,
                "print_functions": [self.print_add_downloads, self.print_table],
            },
        }

    def run(self) -> bool:
        """The main drawing loop."""
        try:
            # outer loop to support screen resize
            while True:
                with ManagedScreen() as screen:
                    logger.debug(f"Created new screen {screen}")
                    self.set_screen(screen)
                    self.frame = 0
                    # break (and re-enter) when screen has been resized
                    while not screen.has_resized():
                        # keep previous sort in memory to know if we have to re-sort the rows
                        # once all events are processed (to avoid useless/redundant sort passes)
                        previous_sort = (self.sort, self.reverse)

                        # we only refresh when explicitly asked for
                        self.refresh = False

                        # process all events before refreshing screen,
                        # otherwise the reactivity is slowed down a lot with fast inputs
                        event = screen.get_event()
                        logger.debug(f"Got event {event}")
                        while event:
                            # avoid crashing the interface if exceptions occur while processing an event
                            try:
                                self.process_event(event)
                            except Exit:
                                logger.debug("Received exit command")
                                return True
                            except Exception as error:  # noqa: BLE001
                                # TODO: display error in status bar
                                logger.exception(error)
                            event = screen.get_event()
                            logger.debug(f"Got event {event}")

                        # time to update data and rows
                        if self.frame == 0:
                            logger.debug("Tick! Updating data and rows")
                            self.update_data()
                            self.update_rows()
                            self.refresh = True

                        # time to refresh the screen
                        if self.refresh:
                            logger.debug("Refresh! Printing text")
                            # sort if needed, unless it was just done at frame 0 when updating
                            if (self.sort, self.reverse) != previous_sort and self.frame != 0:
                                self.sort_data()
                                self.update_rows()

                            # actual printing and screen refresh
                            for print_function in self.state_mapping[self.state]["print_functions"]:
                                print_function()
                            screen.refresh()

                        # sleep and increment frame
                        time.sleep(self.sleep)
                        self.frame = (self.frame + 1) % self.frames
                    logger.debug("Screen has resized")
                    self.post_resize()
        except Exception as error:  # noqa: BLE001
            logger.exception(error)
            return False

    def post_resize(self) -> None:  # noqa: D102
        logger.debug("Running post-resize function")
        logger.debug("Trying to re-apply pywal color theme")
        wal_sequences = Path.home() / ".cache" / "wal" / "sequences"
        try:
            with wal_sequences.open("rb") as fd:
                contents = fd.read()
                sys.stdout.buffer.write(contents)
        except Exception:  # noqa: BLE001,S110
            pass

    def update_select_sort_rows(self) -> None:  # noqa: D102
        self.select_sort_rows = self.columns_order

    def process_event(self, event: KeyboardEvent | MouseEvent) -> None:
        """Process an event.

        For reactivity purpose, this method should not compute expensive stuff, only change the state of the interface,
        changes that will be applied by update_data and update_rows methods.

        Parameters:
            event (KeyboardEvent | MouseEvent): The event to process.
        """
        if isinstance(event, KeyboardEvent):
            self.process_keyboard_event(event)

        elif isinstance(event, MouseEvent):
            self.process_mouse_event(event)

    def process_keyboard_event(self, event: KeyboardEvent) -> None:  # noqa: D102
        self.state_mapping[self.state]["process_keyboard_event"](event)

    def process_keyboard_event_main(self, event: KeyboardEvent) -> None:  # noqa: D102
        if event.key_code in Keys.MOVE_UP:
            if self.focused > 0:
                self.focused -= 1
                logger.debug(f"Move focus up: {self.focused}")

                if self.focused < self.row_offset:
                    self.row_offset = self.focused
                elif self.focused >= self.row_offset + (self.height - 1):
                    # happens when shrinking height
                    self.row_offset = self.focused + 1 - (self.height - 1)
                self.follow = None
                self.refresh = True

        elif event.key_code in Keys.MOVE_DOWN:
            if self.focused < len(self.rows) - 1:
                self.focused += 1
                logger.debug(f"Move focus down: {self.focused}")
                if self.focused - self.row_offset >= (self.height - 1):
                    self.row_offset = self.focused + 1 - (self.height - 1)
                self.follow = None
                self.refresh = True

        elif event.key_code in Keys.MOVE_LEFT:
            if self.x_scroll > 0:
                self.x_scroll = max(0, self.x_scroll - 5)
                self.refresh = True

        elif event.key_code in Keys.MOVE_RIGHT:
            self.x_scroll += 5
            self.refresh = True

        elif event.key_code in Keys.HELP:
            self.state = self.State.HELP
            self.refresh = True

        elif event.key_code in Keys.SETUP:
            pass  # TODO

        elif event.key_code in Keys.TOGGLE_RESUME_PAUSE:
            download = self.data[self.focused]
            if download.is_active or download.is_waiting:
                logger.debug(f"Pausing download {download.gid}")
                download.pause()
            elif download.is_paused:
                logger.debug(f"Resuming download {download.gid}")
                download.resume()

        elif event.key_code in Keys.PRIORITY_UP:
            download = self.data[self.focused]
            if not download.is_active:
                download.move_up()
                self.follow = download

        elif event.key_code in Keys.PRIORITY_DOWN:
            download = self.data[self.focused]
            if not download.is_active:
                download.move_down()
                self.follow = download

        elif event.key_code in Keys.REVERSE_SORT:
            self.reverse = not self.reverse
            self.refresh = True

        elif event.key_code in Keys.NEXT_SORT:
            if self.sort < len(self.columns) - 1:
                self.sort += 1
                self.refresh = True

        elif event.key_code in Keys.PREVIOUS_SORT:
            if self.sort > 0:
                self.sort -= 1
                self.refresh = True

        elif event.key_code in Keys.SELECT_SORT:
            self.state = self.State.SELECT_SORT
            self.side_focused = self.sort
            self.x_offset = self.width_select_sort() + 1
            self.refresh = True

        elif event.key_code in Keys.REMOVE_ASK:
            logger.debug("Triggered removal")
            logger.debug(f"self.focused = {self.focused}")
            logger.debug(f"len(self.data) = {len(self.data)}")
            if self.follow_focused():
                self.state = self.State.REMOVE_ASK
                self.x_offset = self.width_remove_ask() + 1
                if self.last_remove_choice is not None:
                    self.side_focused = self.last_remove_choice
                self.refresh = True
            else:
                logger.debug("Could not focus download")

        elif event.key_code in Keys.TOGGLE_EXPAND_COLLAPSE:  # noqa: SIM114
            pass  # TODO

        elif event.key_code in Keys.TOGGLE_EXPAND_COLLAPSE_ALL:
            pass  # TODO

        elif event.key_code in Keys.AUTOCLEAR:
            self.api.purge()

        elif event.key_code in Keys.FOLLOW_ROW:
            self.follow_focused()

        elif event.key_code in Keys.SEARCH:  # noqa: SIM114
            pass  # TODO

        elif event.key_code in Keys.FILTER:  # noqa: SIM114
            pass  # TODO

        elif event.key_code in Keys.TOGGLE_SELECT:  # noqa: SIM114
            pass  # TODO

        elif event.key_code in Keys.UN_SELECT_ALL:
            pass  # TODO

        elif event.key_code in Keys.MOVE_HOME:
            if self.focused > 0:
                self.focused = 0
                logger.debug(f"Move focus home: {self.focused}")

                if self.focused < self.row_offset:
                    self.row_offset = self.focused
                elif self.focused >= self.row_offset + (self.height - 1):
                    # happens when shrinking height
                    self.row_offset = self.focused + 1 - (self.height - 1)
                self.follow = None
                self.refresh = True

        elif event.key_code in Keys.MOVE_END:
            if self.focused < len(self.rows) - 1:
                self.focused = len(self.rows) - 1
                logger.debug(f"Move focus end: {self.focused}")

                if self.focused - self.row_offset >= (self.height - 1):
                    self.row_offset = self.focused + 1 - (self.height - 1)
                self.follow = None
                self.refresh = True

        elif event.key_code in Keys.MOVE_UP_STEP:
            if self.focused > 0:
                self.focused -= len(self.rows) // 5

                if self.focused < 0:
                    self.focused = 0
                logger.debug(f"Move focus up (step): {self.focused}")

                if self.focused < self.row_offset:
                    self.row_offset = self.focused
                elif self.focused >= self.row_offset + (self.height - 1):
                    # happens when shrinking height
                    self.row_offset = self.focused + 1 - (self.height - 1)

                self.follow = None
                self.refresh = True

        elif event.key_code in Keys.MOVE_DOWN_STEP:
            if self.focused < len(self.rows) - 1:
                self.focused += len(self.rows) // 5

                if self.focused > len(self.rows) - 1:
                    self.focused = len(self.rows) - 1
                logger.debug(f"Move focus down (step): {self.focused}")

                if self.focused - self.row_offset >= (self.height - 1):
                    self.row_offset = self.focused + 1 - (self.height - 1)
                self.follow = None
                self.refresh = True

        elif event.key_code in Keys.TOGGLE_RESUME_PAUSE_ALL:
            stats = self.api.get_stats()
            if stats.num_active:
                self.api.pause_all()
            else:
                self.api.resume_all()

        elif event.key_code in Keys.RETRY:
            download = self.data[self.focused]
            self.api.retry_downloads([download])

        elif event.key_code in Keys.RETRY_ALL:
            downloads = self.data[:]
            self.api.retry_downloads(downloads)

        elif event.key_code in Keys.ADD_DOWNLOADS:
            self.state = self.State.ADD_DOWNLOADS
            self.refresh = True
            self.side_focused = 0
            self.x_offset = self.width

            # build set of copied lines
            copied_lines = set()
            for line in pyperclip.paste().split("\n") + pyperclip.paste(primary=True).split("\n"):
                copied_lines.add(line.strip())
            with contextlib.suppress(KeyError):
                copied_lines.remove("")

            # add lines to download uris
            if copied_lines:
                self.downloads_uris = sorted(copied_lines)

        elif event.key_code in Keys.QUIT:
            raise Exit

    def process_keyboard_event_help(self, event: KeyboardEvent) -> None:  # noqa: ARG002,D102
        self.state = self.State.MAIN
        self.refresh = True

    def process_keyboard_event_setup(self, event: KeyboardEvent) -> None:  # noqa: D102
        pass

    def process_keyboard_event_remove_ask(self, event: KeyboardEvent) -> None:  # noqa: D102
        if event.key_code in Keys.CANCEL:
            logger.debug("Canceling removal")
            self.state = self.State.MAIN
            self.x_offset = 0
            self.refresh = True

        elif event.key_code in Keys.ENTER:
            logger.debug("Validate removal")
            if self.follow:
                self.remove_ask_rows[self.side_focused][1](self.follow)
                self.follow = None
            else:
                logger.debug("No download was targeted, not removing")
            self.last_remove_choice = self.side_focused
            self.state = self.State.MAIN
            self.x_offset = 0

            # force complete refresh
            self.frame = 0

        elif event.key_code in Keys.MOVE_UP:
            if self.side_focused > 0:
                self.side_focused -= 1
                logger.debug(f"Moving side focus up: {self.side_focused}")
                self.refresh = True

        elif event.key_code in Keys.MOVE_DOWN:  # noqa: SIM102
            if self.side_focused < len(self.remove_ask_rows) - 1:
                self.side_focused += 1
                logger.debug(f"Moving side focus down: {self.side_focused}")
                self.refresh = True

    def process_keyboard_event_select_sort(self, event: KeyboardEvent) -> None:  # noqa: D102
        if event.key_code in Keys.CANCEL:
            self.state = self.State.MAIN
            self.x_offset = 0
            self.refresh = True

        elif event.key_code in Keys.ENTER:
            self.sort = self.side_focused
            self.state = self.State.MAIN
            self.x_offset = 0
            self.refresh = True

        elif event.key_code in Keys.MOVE_UP:
            if self.side_focused > 0:
                self.side_focused -= 1
                self.refresh = True

        elif event.key_code in Keys.MOVE_DOWN:  # noqa: SIM102
            if self.side_focused < len(self.select_sort_rows) - 1:
                self.side_focused += 1
                self.refresh = True

    def process_keyboard_event_add_downloads(self, event: KeyboardEvent) -> None:  # noqa: D102
        if event.key_code in Keys.CANCEL:
            self.state = self.State.MAIN
            self.x_offset = 0
            self.refresh = True

        elif event.key_code in Keys.MOVE_UP:
            if self.side_focused > 0:
                self.side_focused -= 1

                if self.side_focused < self.row_offset:
                    self.row_offset = self.side_focused
                elif self.side_focused >= self.row_offset + (self.height - 1):
                    # happens when shrinking height
                    self.row_offset = self.side_focused + 1 - (self.height - 1)
                self.follow = None
                self.refresh = True

        elif event.key_code in Keys.MOVE_DOWN:
            if self.side_focused < len(self.downloads_uris) - 1:
                self.side_focused += 1
                if self.side_focused - self.row_offset >= (self.height - 1):
                    self.row_offset = self.side_focused + 1 - (self.height - 1)
                self.follow = None
                self.refresh = True

        elif event.key_code in Keys.ENTER:
            if self.api.add(self.downloads_uris[self.side_focused]):
                self.downloads_uris.pop(self.side_focused)
                if 0 < self.side_focused > len(self.downloads_uris) - 1:
                    self.side_focused -= 1
                self.refresh = True

        elif event.key_code in Keys.ADD_DOWNLOADS:
            for uri in self.downloads_uris:
                self.api.add(uri)

            self.downloads_uris.clear()
            self.refresh = True

    def process_mouse_event(self, event: MouseEvent) -> None:  # noqa: D102
        self.state_mapping[self.state]["process_mouse_event"](event)

    def process_mouse_event_main(self, event: MouseEvent) -> None:  # noqa: D102
        if event.buttons & MouseEvent.LEFT_CLICK:
            if event.y == 0:
                new_sort = self.get_column_at_x(event.x)
                if new_sort == self.sort:
                    self.reverse = not self.reverse
                else:
                    self.sort = new_sort
            else:
                self.focused = min(event.y - 1 + self.row_offset, len(self.rows) - 1)
            self.refresh = True

        # elif event.buttons & MouseEvent.RIGHT_CLICK:
        #     pass  # TODO: expand/collapse

    def process_mouse_event_help(self, event: MouseEvent) -> None:  # noqa: D102
        pass

    def process_mouse_event_setup(self, event: MouseEvent) -> None:  # noqa: D102
        pass

    def process_mouse_event_remove_ask(self, event: MouseEvent) -> None:  # noqa: D102
        pass

    def process_mouse_event_select_sort(self, event: MouseEvent) -> None:  # noqa: D102
        pass

    def process_mouse_event_add_downloads(self, event: MouseEvent) -> None:  # noqa: D102
        pass

    def width_remove_ask(self) -> int:  # noqa: D102
        return max(len(self.remove_ask_header), max(len(row[0]) for row in self.remove_ask_rows))  # noqa: PLW3301

    def width_select_sort(self) -> int:  # noqa: D102
        return max(len(column_name) for column_name in [*self.columns_order, self.select_sort_header])

    def follow_focused(self) -> bool:  # noqa: D102
        if self.focused < len(self.data):
            self.follow = self.data[self.focused]
            return True
        return False

    def print_add_downloads(self) -> None:  # noqa: D102
        y = self.y_offset
        padding = self.width
        header_string = f"{self.downloads_uris_header:<{padding}}"
        len_header = len(header_string)
        self.screen.print_at(header_string, 0, y, *self.palettes["side_column_header"])
        self.screen.print_at(" ", len_header, y, *self.palettes["default"])
        y += 1
        self.screen.print_at(" " * self.width, 0, y, *self.palettes["ui"])
        separator = "..."

        for i, uri in enumerate(self.downloads_uris):
            y += 1
            palette = (
                self.palettes["side_column_focused_row"] if i == self.side_focused else self.palettes["side_column_row"]
            )
            if len(uri) > self.width:
                # print part of uri string
                uri = f"{uri[:(self.width//2)-len(separator)]} {separator} {uri[-(self.width//2)+len(separator):]}"  # noqa: PLW2901

            self.screen.print_at(uri, 0, y, *palette)
            self.screen.print_at(" ", len(uri), y, *self.palettes["default"])

        for i in range(1, self.height - y):
            self.screen.print_at(" " * (padding + 1), 0, y + i, *self.palettes["ui"])

    def print_help(self) -> None:  # noqa: D102
        version = get_version()
        lines = [
            f"aria2p {version} — (C) 2018-2020 Timothée Mazzucotelli and contributors",
            "Released under the ISC license.",
            "",
        ]

        y = 0
        for line in lines:
            self.screen.print_at(f"{line:<{self.width}}", 0, y, *self.palettes["bright_help"])
            y += 1

        for keys, text in [
            (Keys.HELP, " show this help screen"),
            (Keys.MOVE_UP, " scroll downloads list"),
            (Keys.MOVE_UP_STEP, " scroll downloads list (steps)"),
            (Keys.MOVE_DOWN, " scroll downloads list"),
            (Keys.MOVE_DOWN_STEP, " scroll downloads list (steps)"),
            # not implemented: (Keys.SETUP, " setup"),
            (Keys.TOGGLE_RESUME_PAUSE, " toggle pause/resume"),
            (Keys.PRIORITY_UP, " priority up (-)"),
            (Keys.PRIORITY_DOWN, " priority down (+)"),
            (Keys.REVERSE_SORT, " invert sort order"),
            (Keys.NEXT_SORT, " sort next column"),
            (Keys.PREVIOUS_SORT, " sort previous column"),
            (Keys.SELECT_SORT, " select sort column"),
            (Keys.REMOVE_ASK, " remove download"),
            # not implemented: (Keys.TOGGLE_EXPAND_COLLAPSE, " toggle expand/collapse"),
            # not implemented: (Keys.TOGGLE_EXPAND_COLLAPSE_ALL, " toggle expand/collapse all"),
            (Keys.AUTOCLEAR, " autopurge downloads"),
            (Keys.FOLLOW_ROW, " cursor follows download"),
            # not implemented: (Keys.SEARCH, " name search"),
            # not implemented: (Keys.FILTER, " name filtering"),
            # not implemented: (Keys.TOGGLE_SELECT, " toggle select download"),
            # not implemented: (Keys.UN_SELECT_ALL, " unselect all downloads"),
            (Keys.MOVE_HOME, " move focus to first download"),
            (Keys.MOVE_END, " move focus to last download"),
            (Keys.RETRY, " retry failed download"),
            (Keys.RETRY_ALL, " retry all failed download"),
            (Keys.ADD_DOWNLOADS, " add downloads from clipboard"),
            (Keys.QUIT, " quit"),
        ]:
            self.print_keys(keys, text, y)
            y += 1

        self.screen.print_at(" " * self.width, 0, y, *self.palettes["ui"])
        y += 1
        self.screen.print_at(f"{'Press any key to return.':<{self.width}}", 0, y, *self.palettes["bright_help"])
        y += 1

        for i in range(self.height - y):
            self.screen.print_at(" " * self.width, 0, y + i, *self.palettes["ui"])

    def print_keys(self, keys: list[Key], text: str, y: int) -> None:  # noqa: D102
        self.print_keys_text(" ".join(Keys.names(keys)) + ":", text, y)

    def print_keys_text(self, keys_text: str, text: str, y: int) -> None:  # noqa: D102
        length = 8
        padding = self.width - length
        self.screen.print_at(f"{keys_text:>{length}}", 0, y, *self.palettes["bright_help"])
        self.screen.print_at(f"{text:<{padding}}", length, y, *self.palettes["default"])

    def print_remove_ask_column(self) -> None:  # noqa: D102
        y = self.y_offset
        padding = self.width_remove_ask()
        header_string = f"{self.remove_ask_header:<{padding}}"
        len_header = len(header_string)
        self.screen.print_at(header_string, 0, y, *self.palettes["side_column_header"])
        self.screen.print_at(" ", len_header, y, *self.palettes["default"])
        for i, row in enumerate(self.remove_ask_rows):
            y += 1
            palette = (
                self.palettes["side_column_focused_row"] if i == self.side_focused else self.palettes["side_column_row"]
            )
            row_string = f"{row[0]:<{padding}}"
            len_row = len(row_string)
            self.screen.print_at(row_string, 0, y, *palette)
            self.screen.print_at(" ", len_row, y, *self.palettes["default"])

        for i in range(1, self.height - y):
            self.screen.print_at(" " * (padding + 1), 0, y + i, *self.palettes["ui"])

    def print_select_sort_column(self) -> None:  # noqa: D102
        y = self.y_offset
        padding = self.width_select_sort()
        header_string = f"{self.select_sort_header:<{padding}}"
        len_header = len(header_string)
        self.screen.print_at(header_string, 0, y, *self.palettes["side_column_header"])
        self.screen.print_at(" ", len_header, y, *self.palettes["default"])
        for i, row in enumerate(self.select_sort_rows):
            y += 1
            palette = (
                self.palettes["side_column_focused_row"] if i == self.side_focused else self.palettes["side_column_row"]
            )
            row_string = f"{row:<{padding}}"
            len_row = len(row_string)
            self.screen.print_at(row_string, 0, y, *palette)
            self.screen.print_at(" ", len_row, y, *self.palettes["default"])

        for i in range(1, self.height - y):
            self.screen.print_at(" " * (padding + 1), 0, y + i, *self.palettes["ui"])

    def print_table(self) -> None:  # noqa: D102
        self.print_headers()
        self.print_rows()

    def print_headers(self) -> None:
        """Print the headers (columns names)."""
        self.scroller.set_scroll(self.x_scroll)
        x, y, c = self.x_offset, self.y_offset, 0

        for column_name in self.columns_order:
            column = self.columns[column_name]
            palette = self.palettes["focused_header"] if c == self.sort else self.palettes["header"]

            if column.padding == "100%":
                header_string = f"{column.header}"
                fill_up = " " * max(0, self.width - x - len(header_string))
                written = self.scroller.print_at(header_string, x, y, palette)
                self.scroller.print_at(fill_up, x + written, y, self.palettes["header"])

            else:
                header_string = f"{column.header:{column.padding}} "
                written = self.scroller.print_at(header_string, x, y, palette)

            x += written
            c += 1

    def print_rows(self) -> None:
        """Print the rows."""
        y = self.y_offset + 1
        for row in self.rows[self.row_offset : self.row_offset + self.height]:
            self.scroller.set_scroll(self.x_scroll)
            x = self.x_offset

            for i, column_name in enumerate(self.columns_order):
                column = self.columns[column_name]
                padding = f"<{max(0, self.width - x)}" if column.padding == "100%" else column.padding

                if self.focused == y - self.y_offset - 1 + self.row_offset:
                    palette = self.palettes["focused_row"]
                else:
                    palette = column.get_palette(row[i])
                    if isinstance(palette, str):
                        palette = self.palettes[palette]

                field_string = f"{row[i]:{padding}} "
                written = self.scroller.print_at(field_string, x, y, palette)
                x += written

            y += 1

        for i in range(self.height - y):
            self.screen.print_at(" " * self.width, self.x_offset, y + i, *self.palettes["ui"])

    def get_column_at_x(self, x: int) -> int:
        """For an horizontal position X, return the column index."""
        for i, bound in enumerate(self.bounds):
            if bound[0] <= x <= bound[1]:
                return i
        raise ValueError("clicked outside of boundaries")

    def set_screen(self, screen: Screen) -> None:
        """Set the screen object, its scroller wrapper, width, height, and columns bounds."""
        self.screen = screen
        self.height, self.width = screen.dimensions
        self.scroller = HorizontalScroll(screen)
        self.bounds = []
        for column_name in self.columns_order:
            column = self.columns[column_name]
            if column.padding == "100%":  # last column
                self.bounds.append((self.bounds[-1][1] + 1, self.width))
            else:
                padding = int(column.padding.lstrip("<>=^"))
                if not self.bounds:
                    self.bounds = [(0, padding)]
                else:
                    self.bounds.append((self.bounds[-1][1] + 1, self.bounds[-1][1] + 1 + padding))

    def get_data(self) -> list[Download]:
        """Return a list of objects."""
        return self.api.get_downloads()

    def update_data(self) -> None:
        """Set the interface data and rows contents."""
        try:
            self.data = self.get_data()
            self.sort_data()
        except requests.exceptions.Timeout:
            logger.debug("Request timeout")

    def sort_data(self) -> None:
        """Sort data according to interface state."""
        sort_function = self.columns[self.columns_order[self.sort]].get_sort
        self.data = sorted(self.data, key=sort_function, reverse=self.reverse)

    def update_rows(self) -> None:
        """Update rows contents according to data and interface state."""
        text_getters = [self.columns[c].get_text for c in self.columns_order]
        n_columns = len(self.columns_order)
        self.rows = [tuple(text_getters[i](item) for i in range(n_columns)) for item in self.data]
        if self.follow:
            self.focused = self.data.index(self.follow)



"""Utils module.

This module contains simple utility classes and functions.
"""

from __future__ import annotations

import signal
import sys
import textwrap
from importlib import metadata
from pathlib import Path
from typing import TYPE_CHECKING, Any

from appdirs import user_config_dir
from loguru import logger

if TYPE_CHECKING:
    from datetime import timedelta
    from types import FrameType

if sys.version_info < (3, 11):
    import tomli as tomllib
else:
    import tomllib


class SignalHandler:
    """A helper class to handle signals."""

    def __init__(self, signals: list[str]) -> None:
        """Initialize the object.

        Parameters:
            signals: List of signals names as found in the `signal` module (example: SIGTERM).
        """
        logger.debug("Signal handler: handling signals " + ", ".join(signals))
        self.triggered = False
        for sig in signals:
            try:
                signal.signal(signal.Signals[sig], self.trigger)
            except ValueError as error:
                logger.error(f"Failed to setup signal handler for {sig}: {error}")

    def __bool__(self) -> bool:
        """Return True when one of the given signal was received, False otherwise.

        Returns:
            True when signal received, False otherwise.
        """
        return self.triggered

    def trigger(self, signum: int, frame: FrameType | None) -> None:  # noqa: ARG002
        """Mark this instance as 'triggered' (a specified signal was received).

        Parameters:
            signum: The signal code.
            frame: The signal frame (unused).
        """
        logger.debug(
            f"Signal handler: caught signal {signal.Signals(signum).name} ({signum})",
        )
        self.triggered = True


def human_readable_timedelta(value: timedelta, precision: int = 0) -> str:
    """Return a human-readable time delta as a string.

    Parameters:
        value: The timedelta.
        precision: The precision to use:

            - `0` to display all units
            - `1` to display the biggest unit only
            - `2` to display the first two biggest units only
            - `n` for the first N biggest units, etc.

    Returns:
        A string representing the time delta.
    """
    pieces = []

    if value.days:
        pieces.append(f"{value.days}d")

    seconds = value.seconds

    if seconds >= 3600:  # noqa: PLR2004
        hours = int(seconds / 3600)
        pieces.append(f"{hours}h")
        seconds -= hours * 3600

    if seconds >= 60:  # noqa: PLR2004
        minutes = int(seconds / 60)
        pieces.append(f"{minutes}m")
        seconds -= minutes * 60

    if seconds > 0 or not pieces:
        pieces.append(f"{seconds}s")

    if precision == 0:
        return "".join(pieces)

    return "".join(pieces[:precision])


def human_readable_bytes(value: int, digits: int = 2, delim: str = "", postfix: str = "") -> str:
    """Return a human-readable bytes value as a string.

    Parameters:
        value: The bytes value.
        digits: How many decimal digits to use.
        delim: String to add between value and unit.
        postfix: String to add at the end.

    Returns:
        The human-readable version of the bytes.
    """
    hr_value: float = value
    chosen_unit = "B"
    for unit in ("KiB", "MiB", "GiB", "TiB"):
        if hr_value > 1000:  # noqa: PLR2004
            hr_value /= 1024
            chosen_unit = unit
        else:
            break
    return f"{hr_value:.{digits}f}" + delim + chosen_unit + postfix


def bool_or_value(value: Any) -> Any:
    """Return `True` for `"true"`, `False` for `"false"`, original value otherwise.

    Parameters:
        value: Any kind of value.

    Returns:
        One of these values:

            - `True` for `"true"`
            - `False` for `"false"`
            - Original value otherwise
    """
    if value == "true":
        return True
    if value == "false":
        return False
    return value


def bool_to_str(value: Any) -> Any:
    """Return `"true"` for `True`, `"false"` for `False`, original value otherwise.

    Parameters:
        value: Any kind of value.

    Returns:
        - `"true"` for `True`
        - `"false"` for `False`
        - Original value otherwise
    """
    if value is True:
        return "true"
    if value is False:
        return "false"
    return value


def get_version() -> str:
    """Return the current `aria2p` version.

    Returns:
        The current `aria2p` version.
    """
    try:
        return metadata.version("aria2p")
    except metadata.PackageNotFoundError:
        return "0.0.0"


def load_configuration() -> dict[str, Any]:
    """Return dict from TOML formatted string or file.

    Returns:
        The dict configuration.
    """
    default_config = """
        [key_bindings]
        AUTOCLEAR = "c"
        CANCEL = "esc"
        ENTER = "enter"
        FILTER = ["F4", "\\\\"]
        FOLLOW_ROW = "F"
        HELP = ["F1", "?"]
        MOVE_DOWN = ["down", "j"]
        MOVE_DOWN_STEP = "J"
        MOVE_END = "end"
        MOVE_HOME = "home"
        MOVE_LEFT = ["left", "h"]
        MOVE_RIGHT = ["right", "l"]
        MOVE_UP = ["up", "k"]
        MOVE_UP_STEP = "K"
        NEXT_SORT = ["p", ">"]
        PREVIOUS_SORT = "<"
        PRIORITY_DOWN = ["F8", "d", "]"]
        PRIORITY_UP = ["F7", "u", "["]
        QUIT = ["F10", "q"]
        REMOVE_ASK = ["del", "F9"]
        RETRY = "r"
        RETRY_ALL = "R"
        REVERSE_SORT = "I"
        SEARCH = ["F3", "/"]
        SELECT_SORT = "F6"
        SETUP = "F2"
        TOGGLE_EXPAND_COLLAPSE = "x"
        TOGGLE_EXPAND_COLLAPSE_ALL = "X"
        TOGGLE_RESUME_PAUSE = "space"
        TOGGLE_RESUME_PAUSE_ALL = "P"
        TOGGLE_SELECT = "s"
        UN_SELECT_ALL = "U"
        ADD_DOWNLOADS = "a"

        [colors]
        UI = "WHITE BOLD DEFAULT"
        BRIGHT_HELP = "CYAN BOLD DEFAULT"
        FOCUSED_HEADER = "BLACK NORMAL CYAN"
        FOCUSED_ROW = "BLACK NORMAL CYAN"
        HEADER = "BLACK NORMAL GREEN"
        METADATA = "WHITE UNDERLINE DEFAULT"
        SIDE_COLUMN_FOCUSED_ROW = "DEFAULT NORMAL CYAN"
        SIDE_COLUMN_HEADER = "BLACK NORMAL GREEN"
        SIDE_COLUMN_ROW = "DEFAULT NORMAL DEFAULT"
        STATUS_ACTIVE = "CYAN NORMAL DEFAULT"
        STATUS_COMPLETE = "GREEN NORMAL DEFAULT"
        STATUS_ERROR = "RED BOLD DEFAULT"
        STATUS_PAUSED = "YELLOW NORMAL DEFAULT"
        STATUS_WAITING = "WHITE BOLD DEFAULT"
    """

    config_dict = {}
    config_dict["DEFAULT"] = tomllib.loads(default_config)

    # Check for configuration file
    config_file_path = Path(user_config_dir("aria2p")) / "config.toml"

    if config_file_path.exists():
        try:
            with config_file_path.open("rb") as config_file:
                config_dict["USER"] = tomllib.load(config_file)
        except Exception as error:  # noqa: BLE001
            logger.error(f"Failed to load configuration file: {error}")
    else:
        # Write initial configuration file if it does not exist
        config_file_path.parent.mkdir(parents=True, exist_ok=True)
        with config_file_path.open("w") as fd:
            fd.write(textwrap.dedent(default_config).lstrip("\n"))
    return config_dict


def read_lines(path: str | Path) -> list[str]:
    """Read lines in a file.

    Parameters:
        path: The file path.

    Returns:
        The list of lines.
    """
    return Path(path).read_text().splitlines()



"""Configuration for the pytest test suite."""

from __future__ import annotations

import json
import os
import random
import subprocess
import sys
import time
from pathlib import Path
from typing import Iterator

import psutil
import pytest
import requests

from aria2p import API, Client, enable_logger
from tests import CONFIGS_DIR, SESSIONS_DIR


@pytest.fixture(autouse=True)
def tests_logs(request: pytest.FixtureRequest) -> None:  # noqa: PT004
    # put logs in tests/logs
    log_path = Path("tests") / "logs"

    # tidy logs in subdirectories based on test module and class names
    module = request.module
    class_ = request.cls
    name = request.node.name + ".log"

    if module:
        log_path /= module.__name__.replace("tests.", "")
    if class_:
        log_path /= class_.__name__

    log_path.mkdir(parents=True, exist_ok=True)

    # append last part of the name and enable logger
    log_path /= name
    if log_path.exists():
        log_path.unlink()
    enable_logger(sink=str(log_path), level=os.environ.get("PYTEST_LOG_LEVEL", "TRACE"))


def spawn_and_wait_server(port: int = 8779) -> subprocess.Popen:
    process = subprocess.Popen(
        [  # noqa: S603
            sys.executable,
            "-m",
            "uvicorn",
            "tests.http_server:app",
            "--port",
            str(port),
        ],
        stdout=subprocess.DEVNULL,
        stderr=subprocess.DEVNULL,
    )
    while True:
        try:
            requests.get(f"http://localhost:{port}/1024")  # noqa: S113
        except:  # noqa: E722
            time.sleep(0.1)
        else:
            break
    return process


@pytest.fixture(scope="session", autouse=True)
def http_server(tmp_path_factory: pytest.TempPathFactory, worker_id: str) -> Iterator:
    if worker_id == "master":
        # single worker: just run the HTTP server
        process = spawn_and_wait_server()
        yield process
        process.kill()
        process.wait()
        return

    # get the temp directory shared by all workers
    root_tmp_dir = tmp_path_factory.getbasetemp().parent

    # try to get a lock
    lock = root_tmp_dir / "lock"
    try:
        lock.mkdir(exist_ok=False)
    except FileExistsError:
        yield  # failed, don't run the HTTP server
        return

    # got the lock, run the HTTP server
    process = spawn_and_wait_server()
    yield process
    process.kill()
    process.wait()


class Aria2Server:
    def __init__(
        self,
        tmp_dir: Path,
        port: int,
        config: str | Path | None = None,
        session: str | Path | list[str] | None = None,
        secret: str = "",
    ) -> None:
        """Initialize the server.

        Parameters:
            tmp_dir: Temporary download directory.
            port: Server port.
            config: aria2c configuration file.
            session: aria2c session file.
            secret: Server secret.
        """
        self.tmp_dir = tmp_dir
        self.port = port

        # create the command used to launch an aria2c process
        command = [
            "aria2c",
            f"--dir={self.tmp_dir}",
            "--file-allocation=none",
            "--quiet",
            "--enable-rpc=true",
            f"--rpc-listen-port={self.port}",
        ]
        if config:
            command.append(f"--conf-path={config}")
        else:
            # command.append("--no-conf")
            config = CONFIGS_DIR / "default.conf"
            command.append(f"--conf-path={config}")
        if session:
            if isinstance(session, list):
                session_path = self.tmp_dir / "_session.txt"
                with open(session_path, "w") as stream:
                    stream.write("\n".join(session))
                command.append(f"--input-file={session_path}")
            else:
                session_path = SESSIONS_DIR / session
                if not session_path.exists():
                    raise ValueError(f"no such session: {session}")
                command.append(f"--input-file={session_path}")
        if secret:
            command.append(f"--rpc-secret={secret}")

        self.command = command
        self.process: subprocess.Popen | None = None

        # create the client with port
        self.client = Client(port=self.port, secret=secret, timeout=20)

        # create the API instance
        self.api = API(self.client)

    def __enter__(self) -> Aria2Server:  # noqa: PYI034
        self.start()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):  # noqa: ANN001
        self.destroy(force=True)

    def reach(self, retries: int = 5) -> bool:
        while retries:
            try:
                self.client.list_methods()
            except requests.ConnectionError:
                time.sleep(0.1)
                retries -= 1
            else:
                break
        else:
            return False
        return True

    def start(self) -> None:
        # Make sure we kill any remaining aria2c process using the same port.
        for proc in psutil.process_iter():
            try:
                cmdline = proc.cmdline()
                if "aria2c" in cmdline and f"--rpc-listen-port={self.port}" in cmdline:
                    proc.kill()
                    proc.wait()
                    break
            except psutil.NoSuchProcess:
                pass

        # Make sure we start the new process.
        while True:
            self.process = subprocess.Popen(self.command)  # noqa: S603
            if self.reach(retries=10):
                break
            self.kill()

    def wait(self) -> None:
        if self.process:
            while True:
                try:
                    self.process.wait()
                except subprocess.TimeoutExpired:
                    pass
                else:
                    break

    def terminate(self) -> None:
        if self.process:
            self.process.terminate()
        self.wait()

    def kill(self) -> None:
        if self.process:
            self.process.kill()
        self.wait()

    def rmdir(self, directory: Path | None = None) -> None:
        if directory is None:
            directory = self.tmp_dir
        for item in directory.iterdir():
            if item.is_dir():
                self.rmdir(item)
            else:
                item.unlink()
        directory.rmdir()

    def destroy(self, *, force: bool = False) -> None:
        if force:
            self.kill()
        else:
            self.terminate()
        self.rmdir()


ports_file = Path(".ports.json")


def get_lock() -> None:
    lockdir = Path(".lockdir")
    while True:
        try:
            lockdir.mkdir(exist_ok=False)
        except FileExistsError:
            time.sleep(0.025)
        else:
            break


def release_lock() -> None:
    Path(".lockdir").rmdir()


def get_random_port() -> int:
    return random.randint(15000, 16000)  # noqa: S311


def get_current_ports() -> list[int]:
    try:
        return json.loads(ports_file.read_text())
    except FileNotFoundError:
        return []


def set_current_ports(ports: list[int]) -> None:
    ports_file.write_text(json.dumps(ports))


def reserve_port() -> int:
    get_lock()

    ports = get_current_ports()
    port_number = get_random_port()
    while port_number in ports:
        port_number = get_random_port()
    ports.append(port_number)
    set_current_ports(ports)

    release_lock()
    return port_number


def release_port(port_number: int) -> None:
    get_lock()
    ports = get_current_ports()
    ports.remove(port_number)
    set_current_ports(ports)
    release_lock()


@pytest.fixture()
def port() -> Iterator[int]:
    port_number = reserve_port()
    yield port_number
    release_port(port_number)


@pytest.fixture()
def server(tmp_path: Path, port: int) -> Iterator[Aria2Server]:
    with Aria2Server(tmp_path, port) as server:
        yield server



"""Tests for the `options` module."""

from __future__ import annotations

from aria2p import API, Download, Options


# Test general methods
class TestGeneralMethods:
    def setup_method(self) -> None:
        self.api = API()

    def test_init_method(self) -> None:
        assert Options(self.api, {})

    def test_are_global_method(self) -> None:
        options = Options(self.api, {})
        assert options.are_global

    def test_are_not_global(self) -> None:
        options = Options(self.api, {}, Download(self.api, {}))
        assert not options.are_global

    def test_get_struct_method(self) -> None:
        options = Options(self.api, {"0": 1})
        assert options.get_struct() == {"0": 1}

    def test_get_method(self) -> None:
        options = Options(self.api, {"0": 1})
        assert options.get("0") == 1

    def test_set_method(self) -> None:
        self.api.set_global_options = lambda x: True  # type: ignore[assignment]
        options = Options(self.api, {})
        assert options.set("0", 0)
        assert options.get("0") == "0"

    def test_set_method_failure(self) -> None:
        self.api.set_global_options = lambda x: False  # type: ignore[assignment]
        options = Options(self.api, {"0": "0"})
        assert not options.set("0", "1")
        assert not options.set("1", "1")
        assert options.get("0") == "0"
        assert options.get("1") is None

    def test_set_method_for_download(self) -> None:
        self.api.set_options = lambda x, y: [True]  # type: ignore[assignment]
        options = Options(self.api, {}, Download(self.api, {}))
        assert options.set("0", 0)
        assert options.get("0") == "0"


# Test actual options
class TestOptionsProperties:
    def setup_method(self) -> None:
        self.api = API()
        self.api.set_global_options = lambda x: True  # type: ignore[assignment]
        self.options = Options(self.api, {})

    def test_all_proxy_properties(self) -> None:
        assert self.options.all_proxy is None
        value = "http://user:password@host:2000"
        self.options.all_proxy = value
        assert self.options.all_proxy == value

    def test_all_proxy_passwd_properties(self) -> None:
        assert self.options.all_proxy_passwd is None
        value = ""
        self.options.all_proxy_passwd = value
        assert self.options.all_proxy_passwd == value

    def test_all_proxy_user_properties(self) -> None:
        assert self.options.all_proxy_user is None
        value = ""
        self.options.all_proxy_user = value
        assert self.options.all_proxy_user == value

    def test_allow_overwrite_properties(self) -> None:
        assert self.options.allow_overwrite is None
        value = ""
        self.options.allow_overwrite = value
        assert self.options.allow_overwrite == value

    def test_allow_piece_length_change_properties(self) -> None:
        assert self.options.allow_piece_length_change is None
        value = ""
        self.options.allow_piece_length_change = value
        assert self.options.allow_piece_length_change == value

    def test_always_resume_properties(self) -> None:
        assert self.options.always_resume is None
        value = ""
        self.options.always_resume = value
        assert self.options.always_resume == value

    def test_async_dns_properties(self) -> None:
        assert self.options.async_dns is None
        value = ""
        self.options.async_dns = value
        assert self.options.async_dns == value

    def test_async_dns_server_properties(self) -> None:
        assert self.options.async_dns_server is None
        value = ""
        self.options.async_dns_server = value
        assert self.options.async_dns_server == value

    def test_auto_file_renaming_properties(self) -> None:
        assert self.options.auto_file_renaming is None
        value = ""
        self.options.auto_file_renaming = value
        assert self.options.auto_file_renaming == value

    def test_auto_save_interval_properties(self) -> None:
        assert self.options.auto_save_interval is None
        value = 60
        self.options.auto_save_interval = value
        assert self.options.auto_save_interval == value

    def test_bt_detach_seed_only_properties(self) -> None:
        assert self.options.bt_detach_seed_only is None
        value = ""
        self.options.bt_detach_seed_only = value
        assert self.options.bt_detach_seed_only == value

    def test_bt_enable_hook_after_hash_check_properties(self) -> None:
        assert self.options.bt_enable_hook_after_hash_check is None
        value = ""
        self.options.bt_enable_hook_after_hash_check = value
        assert self.options.bt_enable_hook_after_hash_check == value

    def test_bt_enable_lpd_properties(self) -> None:
        assert self.options.bt_enable_lpd is None
        value = ""
        self.options.bt_enable_lpd = value
        assert self.options.bt_enable_lpd == value

    def test_bt_exclude_tracker_properties(self) -> None:
        assert self.options.bt_exclude_tracker is None
        value = ""
        self.options.bt_exclude_tracker = value
        assert self.options.bt_exclude_tracker == value

    def test_bt_external_ip_properties(self) -> None:
        assert self.options.bt_external_ip is None
        value = ""
        self.options.bt_external_ip = value
        assert self.options.bt_external_ip == value

    def test_bt_force_encryption_properties(self) -> None:
        assert self.options.bt_force_encryption is None
        value = ""
        self.options.bt_force_encryption = value
        assert self.options.bt_force_encryption == value

    def test_bt_hash_check_seed_properties(self) -> None:
        assert self.options.bt_hash_check_seed is None
        value = ""
        self.options.bt_hash_check_seed = value
        assert self.options.bt_hash_check_seed == value

    def test_bt_lpd_interface_properties(self) -> None:
        assert self.options.bt_lpd_interface is None
        value = ""
        self.options.bt_lpd_interface = value
        assert self.options.bt_lpd_interface == value

    def test_bt_max_open_files_properties(self) -> None:
        assert self.options.bt_max_open_files is None
        value = 100
        self.options.bt_max_open_files = value
        assert self.options.bt_max_open_files == value

    def test_bt_max_peers_properties(self) -> None:
        assert self.options.bt_max_peers is None
        value = 50
        self.options.bt_max_peers = value
        assert self.options.bt_max_peers == value

    def test_bt_metadata_only_properties(self) -> None:
        assert self.options.bt_metadata_only is None
        value = ""
        self.options.bt_metadata_only = value
        assert self.options.bt_metadata_only == value

    def test_bt_min_crypto_level_properties(self) -> None:
        assert self.options.bt_min_crypto_level is None
        value = ""
        self.options.bt_min_crypto_level = value
        assert self.options.bt_min_crypto_level == value

    def test_bt_prioritize_piece_properties(self) -> None:
        assert self.options.bt_prioritize_piece is None
        value = ""
        self.options.bt_prioritize_piece = value
        assert self.options.bt_prioritize_piece == value

    def test_bt_remove_unselected_file_properties(self) -> None:
        assert self.options.bt_remove_unselected_file is None
        value = ""
        self.options.bt_remove_unselected_file = value
        assert self.options.bt_remove_unselected_file == value

    def test_bt_request_peer_speed_limit_properties(self) -> None:
        assert self.options.bt_request_peer_speed_limit is None
        value = 256
        self.options.bt_request_peer_speed_limit = value
        assert self.options.bt_request_peer_speed_limit == value

    def test_bt_require_crypto_properties(self) -> None:
        assert self.options.bt_require_crypto is None
        value = ""
        self.options.bt_require_crypto = value
        assert self.options.bt_require_crypto == value

    def test_bt_save_metadata_properties(self) -> None:
        assert self.options.bt_save_metadata is None
        value = ""
        self.options.bt_save_metadata = value
        assert self.options.bt_save_metadata == value

    def test_bt_seed_unverified_properties(self) -> None:
        assert self.options.bt_seed_unverified is None
        value = ""
        self.options.bt_seed_unverified = value
        assert self.options.bt_seed_unverified == value

    def test_bt_stop_timeout_properties(self) -> None:
        assert self.options.bt_stop_timeout is None
        value = 60
        self.options.bt_stop_timeout = value
        assert self.options.bt_stop_timeout == value

    def test_bt_tracker_properties(self) -> None:
        assert self.options.bt_tracker is None
        value = ""
        self.options.bt_tracker = value
        assert self.options.bt_tracker == value

    def test_bt_tracker_connect_timeout_properties(self) -> None:
        assert self.options.bt_tracker_connect_timeout is None
        value = 180
        self.options.bt_tracker_connect_timeout = value
        assert self.options.bt_tracker_connect_timeout == value

    def test_bt_tracker_interval_properties(self) -> None:
        assert self.options.bt_tracker_interval is None
        value = 60
        self.options.bt_tracker_interval = value
        assert self.options.bt_tracker_interval == value

    def test_bt_tracker_timeout_properties(self) -> None:
        assert self.options.bt_tracker_timeout is None
        value = 60
        self.options.bt_tracker_timeout = value
        assert self.options.bt_tracker_timeout == value

    def test_ca_certificate_properties(self) -> None:
        assert self.options.ca_certificate is None
        value = ""
        self.options.ca_certificate = value
        assert self.options.ca_certificate == value

    def test_certificate_properties(self) -> None:
        assert self.options.certificate is None
        value = ""
        self.options.certificate = value
        assert self.options.certificate == value

    def test_check_certificate_properties(self) -> None:
        assert self.options.check_certificate is None
        value = ""
        self.options.check_certificate = value
        assert self.options.check_certificate == value

    def test_check_integrity_properties(self) -> None:
        assert self.options.check_integrity is None
        value = ""
        self.options.check_integrity = value
        assert self.options.check_integrity == value

    def test_checksum_properties(self) -> None:
        assert self.options.checksum is None
        value = ""
        self.options.checksum = value
        assert self.options.checksum == value

    def test_conditional_get_properties(self) -> None:
        assert self.options.conditional_get is None
        value = ""
        self.options.conditional_get = value
        assert self.options.conditional_get == value

    def test_conf_path_properties(self) -> None:
        assert self.options.conf_path is None
        value = ""
        self.options.conf_path = value
        assert self.options.conf_path == value

    def test_connect_timeout_properties(self) -> None:
        assert self.options.connect_timeout is None
        value = 60
        self.options.connect_timeout = value
        assert self.options.connect_timeout == value

    def test_console_log_level_properties(self) -> None:
        assert self.options.console_log_level is None
        value = ""
        self.options.console_log_level = value
        assert self.options.console_log_level == value

    def test_continue_downloads_properties(self) -> None:
        assert self.options.continue_downloads is None
        value = ""
        self.options.continue_downloads = value
        assert self.options.continue_downloads == value

    def test_daemon_properties(self) -> None:
        assert self.options.daemon is None
        value = ""
        self.options.daemon = value
        assert self.options.daemon == value

    def test_deferred_input_properties(self) -> None:
        assert self.options.deferred_input is None
        value = ""
        self.options.deferred_input = value
        assert self.options.deferred_input == value

    def test_dht_entry_point_properties(self) -> None:
        assert self.options.dht_entry_point is None
        value = ""
        self.options.dht_entry_point = value
        assert self.options.dht_entry_point == value

    def test_dht_entry_point6_properties(self) -> None:
        assert self.options.dht_entry_point6 is None
        value = ""
        self.options.dht_entry_point6 = value
        assert self.options.dht_entry_point6 == value

    def test_dht_file_path_properties(self) -> None:
        assert self.options.dht_file_path is None
        value = ""
        self.options.dht_file_path = value
        assert self.options.dht_file_path == value

    def test_dht_file_path6_properties(self) -> None:
        assert self.options.dht_file_path6 is None
        value = ""
        self.options.dht_file_path6 = value
        assert self.options.dht_file_path6 == value

    def test_dht_listen_addr6_properties(self) -> None:
        assert self.options.dht_listen_addr6 is None
        value = ""
        self.options.dht_listen_addr6 = value
        assert self.options.dht_listen_addr6 == value

    def test_dht_listen_port_properties(self) -> None:
        assert self.options.dht_listen_port is None
        value = ""
        self.options.dht_listen_port = value
        assert self.options.dht_listen_port == value

    def test_dht_message_timeout_properties(self) -> None:
        assert self.options.dht_message_timeout is None
        value = 30
        self.options.dht_message_timeout = value
        assert self.options.dht_message_timeout == value

    def test_dir_properties(self) -> None:
        assert self.options.dir is None
        value = "/some/random/dir"
        self.options.dir = value
        assert self.options.dir == value

    def test_disable_ipv6_properties(self) -> None:
        assert self.options.disable_ipv6 is None
        value = ""
        self.options.disable_ipv6 = value
        assert self.options.disable_ipv6 == value

    def test_disk_cache_properties(self) -> None:
        assert self.options.disk_cache is None
        value = 2048
        self.options.disk_cache = value
        assert self.options.disk_cache == value

    def test_download_result_properties(self) -> None:
        assert self.options.download_result is None
        value = ""
        self.options.download_result = value
        assert self.options.download_result == value

    def test_dry_run_properties(self) -> None:
        assert self.options.dry_run is None
        value = ""
        self.options.dry_run = value
        assert self.options.dry_run == value

    def test_dscp_properties(self) -> None:
        assert self.options.dscp is None
        value = ""
        self.options.dscp = value
        assert self.options.dscp == value

    def test_enable_color_properties(self) -> None:
        assert self.options.enable_color is None
        value = ""
        self.options.enable_color = value
        assert self.options.enable_color == value

    def test_enable_dht_properties(self) -> None:
        assert self.options.enable_dht is None
        value = ""
        self.options.enable_dht = value
        assert self.options.enable_dht == value

    def test_enable_dht6_properties(self) -> None:
        assert self.options.enable_dht6 is None
        value = ""
        self.options.enable_dht6 = value
        assert self.options.enable_dht6 == value

    def test_enable_http_keep_alive_properties(self) -> None:
        assert self.options.enable_http_keep_alive is None
        value = ""
        self.options.enable_http_keep_alive = value
        assert self.options.enable_http_keep_alive == value

    def test_enable_http_pipelining_properties(self) -> None:
        assert self.options.enable_http_pipelining is None
        value = ""
        self.options.enable_http_pipelining = value
        assert self.options.enable_http_pipelining == value

    def test_enable_mmap_properties(self) -> None:
        assert self.options.enable_mmap is None
        value = ""
        self.options.enable_mmap = value
        assert self.options.enable_mmap == value

    def test_enable_peer_exchange_properties(self) -> None:
        assert self.options.enable_peer_exchange is None
        value = ""
        self.options.enable_peer_exchange = value
        assert self.options.enable_peer_exchange == value

    def test_enable_rpc_properties(self) -> None:
        assert self.options.enable_rpc is None
        value = ""
        self.options.enable_rpc = value
        assert self.options.enable_rpc == value

    def test_event_poll_properties(self) -> None:
        assert self.options.event_poll is None
        value = ""
        self.options.event_poll = value
        assert self.options.event_poll == value

    def test_file_allocation_properties(self) -> None:
        assert self.options.file_allocation is None
        value = ""
        self.options.file_allocation = value
        assert self.options.file_allocation == value

    def test_follow_metalink_properties(self) -> None:
        assert self.options.follow_metalink is None
        value = ""
        self.options.follow_metalink = value
        assert self.options.follow_metalink == value

    def test_follow_torrent_properties(self) -> None:
        assert self.options.follow_torrent is None
        value = ""
        self.options.follow_torrent = value
        assert self.options.follow_torrent == value

    def test_force_save_properties(self) -> None:
        assert self.options.force_save is None
        value = ""
        self.options.force_save = value
        assert self.options.force_save == value

    def test_force_sequential_properties(self) -> None:
        assert self.options.force_sequential is None
        value = ""
        self.options.force_sequential = value
        assert self.options.force_sequential == value

    def test_ftp_passwd_properties(self) -> None:
        assert self.options.ftp_passwd is None
        value = ""
        self.options.ftp_passwd = value
        assert self.options.ftp_passwd == value

    def test_ftp_pasv_properties(self) -> None:
        assert self.options.ftp_pasv is None
        value = ""
        self.options.ftp_pasv = value
        assert self.options.ftp_pasv == value

    def test_ftp_proxy_properties(self) -> None:
        assert self.options.ftp_proxy is None
        value = ""
        self.options.ftp_proxy = value
        assert self.options.ftp_proxy == value

    def test_ftp_proxy_passwd_properties(self) -> None:
        assert self.options.ftp_proxy_passwd is None
        value = ""
        self.options.ftp_proxy_passwd = value
        assert self.options.ftp_proxy_passwd == value

    def test_ftp_proxy_user_properties(self) -> None:
        assert self.options.ftp_proxy_user is None
        value = ""
        self.options.ftp_proxy_user = value
        assert self.options.ftp_proxy_user == value

    def test_ftp_reuse_connection_properties(self) -> None:
        assert self.options.ftp_reuse_connection is None
        value = ""
        self.options.ftp_reuse_connection = value
        assert self.options.ftp_reuse_connection == value

    def test_ftp_type_properties(self) -> None:
        assert self.options.ftp_type is None
        value = ""
        self.options.ftp_type = value
        assert self.options.ftp_type == value

    def test_ftp_user_properties(self) -> None:
        assert self.options.ftp_user is None
        value = ""
        self.options.ftp_user = value
        assert self.options.ftp_user == value

    def test_gid_properties(self) -> None:
        assert self.options.gid is None
        value = ""
        self.options.gid = value
        assert self.options.gid == value

    def test_hash_check_only_properties(self) -> None:
        assert self.options.hash_check_only is None
        value = ""
        self.options.hash_check_only = value
        assert self.options.hash_check_only == value

    def test_header_properties(self) -> None:
        assert self.options.header is None
        value = ""
        self.options.header = value
        assert self.options.header == value

    def test_http_accept_gzip_properties(self) -> None:
        assert self.options.http_accept_gzip is None
        value = ""
        self.options.http_accept_gzip = value
        assert self.options.http_accept_gzip == value

    def test_http_auth_challenge_properties(self) -> None:
        assert self.options.http_auth_challenge is None
        value = ""
        self.options.http_auth_challenge = value
        assert self.options.http_auth_challenge == value

    def test_http_no_cache_properties(self) -> None:
        assert self.options.http_no_cache is None
        value = ""
        self.options.http_no_cache = value
        assert self.options.http_no_cache == value

    def test_http_passwd_properties(self) -> None:
        assert self.options.http_passwd is None
        value = ""
        self.options.http_passwd = value
        assert self.options.http_passwd == value

    def test_http_proxy_properties(self) -> None:
        assert self.options.http_proxy is None
        value = ""
        self.options.http_proxy = value
        assert self.options.http_proxy == value

    def test_http_proxy_passwd_properties(self) -> None:
        assert self.options.http_proxy_passwd is None
        value = ""
        self.options.http_proxy_passwd = value
        assert self.options.http_proxy_passwd == value

    def test_http_proxy_user_properties(self) -> None:
        assert self.options.http_proxy_user is None
        value = ""
        self.options.http_proxy_user = value
        assert self.options.http_proxy_user == value

    def test_http_user_properties(self) -> None:
        assert self.options.http_user is None
        value = ""
        self.options.http_user = value
        assert self.options.http_user == value

    def test_https_proxy_properties(self) -> None:
        assert self.options.https_proxy is None
        value = ""
        self.options.https_proxy = value
        assert self.options.https_proxy == value

    def test_https_proxy_passwd_properties(self) -> None:
        assert self.options.https_proxy_passwd is None
        value = ""
        self.options.https_proxy_passwd = value
        assert self.options.https_proxy_passwd == value

    def test_https_proxy_user_properties(self) -> None:
        assert self.options.https_proxy_user is None
        value = ""
        self.options.https_proxy_user = value
        assert self.options.https_proxy_user == value

    def test_human_readable_properties(self) -> None:
        assert self.options.human_readable is None
        value = ""
        self.options.human_readable = value
        assert self.options.human_readable == value

    def test_index_out_properties(self) -> None:
        assert self.options.index_out is None
        value = ""
        self.options.index_out = value
        assert self.options.index_out == value

    def test_input_file_properties(self) -> None:
        assert self.options.input_file is None
        value = ""
        self.options.input_file = value
        assert self.options.input_file == value

    def test_interface_properties(self) -> None:
        assert self.options.interface is None
        value = ""
        self.options.interface = value
        assert self.options.interface == value

    def test_keep_unfinished_download_result_properties(self) -> None:
        assert self.options.keep_unfinished_download_result is None
        value = ""
        self.options.keep_unfinished_download_result = value
        assert self.options.keep_unfinished_download_result == value

    def test_listen_port_properties(self) -> None:
        assert self.options.listen_port is None
        value = ""
        self.options.listen_port = value
        assert self.options.listen_port == value

    def test_load_cookies(self) -> None:
        assert self.options.load_cookies is None
        value = ""
        self.options.load_cookies = value
        assert self.options.load_cookies == value

    def test_log_properties(self) -> None:
        assert self.options.log is None
        value = ""
        self.options.log = value
        assert self.options.log == value

    def test_log_level_properties(self) -> None:
        assert self.options.log_level is None
        value = ""
        self.options.log_level = value
        assert self.options.log_level == value

    def test_lowest_speed_limit_properties(self) -> None:
        assert self.options.lowest_speed_limit is None
        value = 128
        self.options.lowest_speed_limit = value
        assert self.options.lowest_speed_limit == value

    def test_max_concurrent_downloads_properties(self) -> None:
        assert self.options.max_concurrent_downloads is None
        value = 10
        self.options.max_concurrent_downloads = value
        assert self.options.max_concurrent_downloads == value

    def test_max_connection_per_server_properties(self) -> None:
        assert self.options.max_connection_per_server is None
        value = 20
        self.options.max_connection_per_server = value
        assert self.options.max_connection_per_server == value

    def test_max_download_limit_properties(self) -> None:
        assert self.options.max_download_limit is None
        value = 100
        self.options.max_download_limit = value
        assert self.options.max_download_limit == value

    def test_max_download_result_properties(self) -> None:
        assert self.options.max_download_result is None
        value = 50
        self.options.max_download_result = value
        assert self.options.max_download_result == value

    def test_max_file_not_found_properties(self) -> None:
        assert self.options.max_file_not_found is None
        value = 10
        self.options.max_file_not_found = value
        assert self.options.max_file_not_found == value

    def test_max_mmap_limit_properties(self) -> None:
        assert self.options.max_mmap_limit is None
        value = 10
        self.options.max_mmap_limit = value
        assert self.options.max_mmap_limit == value

    def test_max_overall_download_limit_properties(self) -> None:
        assert self.options.max_overall_download_limit is None
        value = 1000
        self.options.max_overall_download_limit = value
        assert self.options.max_overall_download_limit == value

    def test_max_overall_upload_limit_properties(self) -> None:
        assert self.options.max_overall_upload_limit is None
        value = 1000
        self.options.max_overall_upload_limit = value
        assert self.options.max_overall_upload_limit == value

    def test_max_resume_failure_tries_properties(self) -> None:
        assert self.options.max_resume_failure_tries is None
        value = 100
        self.options.max_resume_failure_tries = value
        assert self.options.max_resume_failure_tries == value

    def test_max_tries_properties(self) -> None:
        assert self.options.max_tries is None
        value = 1
        self.options.max_tries = value
        assert self.options.max_tries == value

    def test_max_upload_limit_properties(self) -> None:
        assert self.options.max_upload_limit is None
        value = 100
        self.options.max_upload_limit = value
        assert self.options.max_upload_limit == value

    def test_metalink_base_uri_properties(self) -> None:
        assert self.options.metalink_base_uri is None
        value = ""
        self.options.metalink_base_uri = value
        assert self.options.metalink_base_uri == value

    def test_metalink_enable_unique_protocol_properties(self) -> None:
        assert self.options.metalink_enable_unique_protocol is None
        value = ""
        self.options.metalink_enable_unique_protocol = value
        assert self.options.metalink_enable_unique_protocol == value

    def test_metalink_file_properties(self) -> None:
        assert self.options.metalink_file is None
        value = ""
        self.options.metalink_file = value
        assert self.options.metalink_file == value

    def test_metalink_language_properties(self) -> None:
        assert self.options.metalink_language is None
        value = ""
        self.options.metalink_language = value
        assert self.options.metalink_language == value

    def test_metalink_location_properties(self) -> None:
        assert self.options.metalink_location is None
        value = ""
        self.options.metalink_location = value
        assert self.options.metalink_location == value

    def test_metalink_os_properties(self) -> None:
        assert self.options.metalink_os is None
        value = ""
        self.options.metalink_os = value
        assert self.options.metalink_os == value

    def test_metalink_preferred_protocol_properties(self) -> None:
        assert self.options.metalink_preferred_protocol is None
        value = ""
        self.options.metalink_preferred_protocol = value
        assert self.options.metalink_preferred_protocol == value

    def test_metalink_version_properties(self) -> None:
        assert self.options.metalink_version is None
        value = ""
        self.options.metalink_version = value
        assert self.options.metalink_version == value

    def test_min_split_size_properties(self) -> None:
        assert self.options.min_split_size is None
        value = 2048
        self.options.min_split_size = value
        assert self.options.min_split_size == value

    def test_min_tls_version_properties(self) -> None:
        assert self.options.min_tls_version is None
        value = ""
        self.options.min_tls_version = value
        assert self.options.min_tls_version == value

    def test_multiple_interface_properties(self) -> None:
        assert self.options.multiple_interface is None
        value = ""
        self.options.multiple_interface = value
        assert self.options.multiple_interface == value

    def test_netrc_path_properties(self) -> None:
        assert self.options.netrc_path is None
        value = ""
        self.options.netrc_path = value
        assert self.options.netrc_path == value

    def test_no_conf_properties(self) -> None:
        assert self.options.no_conf is None
        value = ""
        self.options.no_conf = value
        assert self.options.no_conf == value

    def test_no_file_allocation_limit_properties(self) -> None:
        assert self.options.no_file_allocation_limit is None
        value = 200
        self.options.no_file_allocation_limit = value
        assert self.options.no_file_allocation_limit == value

    def test_no_netrc_properties(self) -> None:
        assert self.options.no_netrc is None
        value = ""
        self.options.no_netrc = value
        assert self.options.no_netrc == value

    def test_no_proxy_properties(self) -> None:
        assert self.options.no_proxy is None
        value = ""
        self.options.no_proxy = value
        assert self.options.no_proxy == value

    def test_on_bt_download_complete_properties(self) -> None:
        assert self.options.on_bt_download_complete is None
        value = ""
        self.options.on_bt_download_complete = value
        assert self.options.on_bt_download_complete == value

    def test_on_download_complete_properties(self) -> None:
        assert self.options.on_download_complete is None
        value = ""
        self.options.on_download_complete = value
        assert self.options.on_download_complete == value

    def test_on_download_error_properties(self) -> None:
        assert self.options.on_download_error is None
        value = ""
        self.options.on_download_error = value
        assert self.options.on_download_error == value

    def test_on_download_pause_properties(self) -> None:
        assert self.options.on_download_pause is None
        value = ""
        self.options.on_download_pause = value
        assert self.options.on_download_pause == value

    def test_on_download_start_properties(self) -> None:
        assert self.options.on_download_start is None
        value = ""
        self.options.on_download_start = value
        assert self.options.on_download_start == value

    def test_on_download_stop_properties(self) -> None:
        assert self.options.on_download_stop is None
        value = ""
        self.options.on_download_stop = value
        assert self.options.on_download_stop == value

    def test_optimize_concurrent_downloads_properties(self) -> None:
        assert self.options.optimize_concurrent_downloads is None
        value = ""
        self.options.optimize_concurrent_downloads = value
        assert self.options.optimize_concurrent_downloads == value

    def test_out_properties(self) -> None:
        assert self.options.out is None
        value = ""
        self.options.out = value
        assert self.options.out == value

    def test_parameterized_uri_properties(self) -> None:
        assert self.options.parameterized_uri is None
        value = ""
        self.options.parameterized_uri = value
        assert self.options.parameterized_uri == value

    def test_pause_properties(self) -> None:
        assert self.options.pause is None
        value = ""
        self.options.pause = value
        assert self.options.pause == value

    def test_pause_metadata_properties(self) -> None:
        assert self.options.pause_metadata is None
        value = ""
        self.options.pause_metadata = value
        assert self.options.pause_metadata == value

    def test_peer_id_prefix_properties(self) -> None:
        assert self.options.peer_id_prefix is None
        value = ""
        self.options.peer_id_prefix = value
        assert self.options.peer_id_prefix == value

    def test_piece_length_properties(self) -> None:
        assert self.options.piece_length is None
        value = ""
        self.options.piece_length = value
        assert self.options.piece_length == value

    def test_private_key_properties(self) -> None:
        assert self.options.private_key is None
        value = ""
        self.options.private_key = value
        assert self.options.private_key == value

    def test_proxy_method_properties(self) -> None:
        assert self.options.proxy_method is None
        value = ""
        self.options.proxy_method = value
        assert self.options.proxy_method == value

    def test_quiet_properties(self) -> None:
        assert self.options.quiet is None
        value = ""
        self.options.quiet = value
        assert self.options.quiet == value

    def test_realtime_chunk_checksum_properties(self) -> None:
        assert self.options.realtime_chunk_checksum is None
        value = ""
        self.options.realtime_chunk_checksum = value
        assert self.options.realtime_chunk_checksum == value

    def test_referer_properties(self) -> None:
        assert self.options.referer is None
        value = ""
        self.options.referer = value
        assert self.options.referer == value

    def test_remote_time_properties(self) -> None:
        assert self.options.remote_time is None
        value = ""
        self.options.remote_time = value
        assert self.options.remote_time == value

    def test_remove_control_file_properties(self) -> None:
        assert self.options.remove_control_file is None
        value = ""
        self.options.remove_control_file = value
        assert self.options.remove_control_file == value

    def test_retry_wait_properties(self) -> None:
        assert self.options.retry_wait is None
        value = 2
        self.options.retry_wait = value
        assert self.options.retry_wait == value

    def test_reuse_uri_properties(self) -> None:
        assert self.options.reuse_uri is None
        value = ""
        self.options.reuse_uri = value
        assert self.options.reuse_uri == value

    def test_rlimit_nofile(self) -> None:
        assert self.options.rlimit_nofile is None
        value = 10
        self.options.rlimit_nofile = value
        assert self.options.rlimit_nofile == value

    def test_rpc_allow_origin_all_properties(self) -> None:
        assert self.options.rpc_allow_origin_all is None
        value = ""
        self.options.rpc_allow_origin_all = value
        assert self.options.rpc_allow_origin_all == value

    def test_rpc_certificate_properties(self) -> None:
        assert self.options.rpc_certificate is None
        value = ""
        self.options.rpc_certificate = value
        assert self.options.rpc_certificate == value

    def test_rpc_listen_all_properties(self) -> None:
        assert self.options.rpc_listen_all is None
        value = True
        self.options.rpc_listen_all = value
        assert self.options.rpc_listen_all == value

    def test_rpc_listen_port_properties(self) -> None:
        assert self.options.rpc_listen_port is None
        value = 6801
        self.options.rpc_listen_port = value
        assert self.options.rpc_listen_port == value

    def test_rpc_max_request_size_properties(self) -> None:
        assert self.options.rpc_max_request_size is None
        value = ""
        self.options.rpc_max_request_size = value
        assert self.options.rpc_max_request_size == value

    def test_rpc_passwd_properties(self) -> None:
        assert self.options.rpc_passwd is None
        value = ""
        self.options.rpc_passwd = value
        assert self.options.rpc_passwd == value

    def test_rpc_private_key_properties(self) -> None:
        assert self.options.rpc_private_key is None
        value = ""
        self.options.rpc_private_key = value
        assert self.options.rpc_private_key == value

    def test_rpc_save_upload_metadata_properties(self) -> None:
        assert self.options.rpc_save_upload_metadata is None
        value = ""
        self.options.rpc_save_upload_metadata = value
        assert self.options.rpc_save_upload_metadata == value

    def test_rpc_secret(self) -> None:
        assert self.options.rpc_secret is None
        value = "secret"
        self.options.rpc_secret = value
        assert self.options.rpc_secret == value

    def test_rpc_secure_properties(self) -> None:
        assert self.options.rpc_secure is None
        value = ""
        self.options.rpc_secure = value
        assert self.options.rpc_secure == value

    def test_rpc_user_properties(self) -> None:
        assert self.options.rpc_user is None
        value = ""
        self.options.rpc_user = value
        assert self.options.rpc_user == value

    def test_save_cookies_properties(self) -> None:
        assert self.options.save_cookies is None
        value = ""
        self.options.save_cookies = value
        assert self.options.save_cookies == value

    def test_save_not_found_properties(self) -> None:
        assert self.options.save_not_found is None
        value = ""
        self.options.save_not_found = value
        assert self.options.save_not_found == value

    def test_save_session_properties(self) -> None:
        assert self.options.save_session is None
        value = ""
        self.options.save_session = value
        assert self.options.save_session == value

    def test_save_session_interval_properties(self) -> None:
        assert self.options.save_session_interval is None
        value = 120
        self.options.save_session_interval = value
        assert self.options.save_session_interval == value

    def test_seed_ratio_properties(self) -> None:
        assert self.options.seed_ratio is None
        value = ""
        self.options.seed_ratio = value
        assert self.options.seed_ratio == value

    def test_seed_time_properties(self) -> None:
        assert self.options.seed_time is None
        value = ""
        self.options.seed_time = value
        assert self.options.seed_time == value

    def test_select_file_properties(self) -> None:
        assert self.options.select_file is None
        value = ""
        self.options.select_file = value
        assert self.options.select_file == value

    def test_server_stat_if_properties(self) -> None:
        assert self.options.server_stat_if is None
        value = ""
        self.options.server_stat_if = value
        assert self.options.server_stat_if == value

    def test_server_stat_of_properties(self) -> None:
        assert self.options.server_stat_of is None
        value = ""
        self.options.server_stat_of = value
        assert self.options.server_stat_of == value

    def test_server_stat_timeout_properties(self) -> None:
        assert self.options.server_stat_timeout is None
        value = 60
        self.options.server_stat_timeout = value
        assert self.options.server_stat_timeout == value

    def test_show_console_readout_properties(self) -> None:
        assert self.options.show_console_readout is None
        value = ""
        self.options.show_console_readout = value
        assert self.options.show_console_readout == value

    def test_show_files_properties(self) -> None:
        assert self.options.show_files is None
        value = ""
        self.options.show_files = value
        assert self.options.show_files == value

    def test_socket_recv_buffer_size_properties(self) -> None:
        assert self.options.socket_recv_buffer_size is None
        value = 256
        self.options.socket_recv_buffer_size = value
        assert self.options.socket_recv_buffer_size == value

    def test_split_properties(self) -> None:
        assert self.options.split is None
        value = 5
        self.options.split = value
        assert self.options.split == value

    def test_ssh_host_key_md_properties(self) -> None:
        assert self.options.ssh_host_key_md is None
        value = ""
        self.options.ssh_host_key_md = value
        assert self.options.ssh_host_key_md == value

    def test_stderr_properties(self) -> None:
        assert self.options.stderr is None
        value = ""
        self.options.stderr = value
        assert self.options.stderr == value

    def test_stop(self) -> None:
        assert self.options.stop is None
        value = 1000
        self.options.stop = value
        assert self.options.stop == value

    def test_stop_with_process_properties(self) -> None:
        assert self.options.stop_with_process is None
        value = 15050
        self.options.stop_with_process = value
        assert self.options.stop_with_process == value

    def test_stream_piece_selector_properties(self) -> None:
        assert self.options.stream_piece_selector is None
        value = ""
        self.options.stream_piece_selector = value
        assert self.options.stream_piece_selector == value

    def test_summary_interval_properties(self) -> None:
        assert self.options.summary_interval is None
        value = 120
        self.options.summary_interval = value
        assert self.options.summary_interval == value

    def test_timeout_properties(self) -> None:
        assert self.options.timeout is None
        value = 60
        self.options.timeout = value
        assert self.options.timeout == value

    def test_torrent_file_properties(self) -> None:
        assert self.options.torrent_file is None
        value = ""
        self.options.torrent_file = value
        assert self.options.torrent_file == value

    def test_truncate_console_readout_properties(self) -> None:
        assert self.options.truncate_console_readout is None
        value = ""
        self.options.truncate_console_readout = value
        assert self.options.truncate_console_readout == value

    def test_uri_selector_properties(self) -> None:
        assert self.options.uri_selector is None
        value = ""
        self.options.uri_selector = value
        assert self.options.uri_selector == value

    def test_use_head_properties(self) -> None:
        assert self.options.use_head is None
        value = ""
        self.options.use_head = value
        assert self.options.use_head == value

    def test_user_agent_properties(self) -> None:
        assert self.options.user_agent is None
        value = ""
        self.options.user_agent = value
        assert self.options.user_agent == value



"""Tests for the `stats` module."""

from __future__ import annotations

import pytest

from aria2p import Stats


@pytest.fixture(scope="module")
def stats() -> Stats:
    """Provide a Stats instance.

    Returns:
        A Stats instance.
    """
    return Stats(
        {
            "downloadSpeed": "0",
            "numActive": "0",
            "numStopped": "0",
            "numStoppedTotal": "0",
            "numWaiting": "0",
            "uploadSpeed": "0",
        },
    )


def test_download_speed(stats: Stats) -> None:
    """Test the `download_speed` property.

    Parameters:
        stats: A Stats instance.
    """
    assert stats.download_speed == 0


def test_download_speed_string(stats: Stats) -> None:
    """Test the `download_speed_string` method.

    Parameters:
        stats: A Stats instance.
    """
    assert stats.download_speed_string() == "0.00 B/s"


def test_download_speed_string_not_human_readable(stats: Stats) -> None:
    """Test the `download_speed_string` method with human readable option.

    Parameters:
        stats: A Stats instance.
    """
    assert stats.download_speed_string(human_readable=False) == "0 B/s"


def test_upload_speed(stats: Stats) -> None:
    """Test the `upload_speed` property.

    Parameters:
        stats: A Stats instance.
    """
    assert stats.upload_speed == 0


def test_upload_speed_string(stats: Stats) -> None:
    """Test the `upload_speed_string` method.

    Parameters:
        stats: A Stats instance.
    """
    assert stats.upload_speed_string() == "0.00 B/s"


def test_upload_speed_string_not_human_readable(stats: Stats) -> None:
    """Test the `upload_speed_string` method with human readable option.

    Parameters:
        stats: A Stats instance.
    """
    assert stats.upload_speed_string(human_readable=False) == "0 B/s"


def test_num_active(stats: Stats) -> None:
    """Test the `numnum_active_waiting` property.

    Parameters:
        stats: A Stats instance.
    """
    assert stats.num_active == 0


def test_num_stopped(stats: Stats) -> None:
    """Test the `num_stopped` property.

    Parameters:
        stats: A Stats instance.
    """
    assert stats.num_stopped == 0


def test_num_stopped_total(stats: Stats) -> None:
    """Test the `num_stopped_total` property.

    Parameters:
        stats: A Stats instance.
    """
    assert stats.num_stopped_total == 0


def test_num_waiting(stats: Stats) -> None:
    """Test the `num_waiting` property.

    Parameters:
        stats: A Stats instance.
    """
    assert stats.num_waiting == 0



"""Tests for the `api` module."""

from __future__ import annotations

import threading
import time
from typing import TYPE_CHECKING

import pytest

from aria2p import API, Client, ClientException, Download
from tests import BUNSENLABS_MAGNET, BUNSENLABS_TORRENT, CONFIGS_DIR, DEBIAN_METALINK, INPUT_FILES, XUBUNTU_MIRRORS
from tests.conftest import Aria2Server

if TYPE_CHECKING:
    from pathlib import Path


def test_add_magnet_method(server: Aria2Server) -> None:
    assert server.api.add_magnet(BUNSENLABS_MAGNET)


def test_add_metalink_method(server: Aria2Server) -> None:
    assert server.api.add_metalink(DEBIAN_METALINK)


def test_add_torrent_method(server: Aria2Server) -> None:
    assert server.api.add_torrent(BUNSENLABS_TORRENT)


def test_add_uris_method(server: Aria2Server) -> None:
    assert server.api.add_uris(XUBUNTU_MIRRORS)


def test_get_download_method(tmp_path: Path, port: int) -> None:
    with Aria2Server(tmp_path, port, session="1-dl-paused.txt") as server:
        assert server.api.get_download("0000000000000001")  # == server.api.get_downloads()[0].gid


def test_get_downloads_method(tmp_path: Path, port: int) -> None:
    with Aria2Server(tmp_path, port, session="2-dls.txt") as server:
        downloads = server.api.get_downloads()
        assert len(downloads) == 2
        assert isinstance(downloads[0], Download)
        assert downloads[0].gid == "0000000000000001"


def test_get_global_options_method(tmp_path: Path, port: int) -> None:
    with Aria2Server(tmp_path, port, config=CONFIGS_DIR / "max-5-dls.conf") as server:
        options = server.api.get_global_options()
        assert options.download is None
        assert options.max_concurrent_downloads == 5


def test_get_options_method(tmp_path: Path, port: int) -> None:
    with Aria2Server(tmp_path, port, session="max-dl-limit-10000.txt") as server:
        downloads = server.api.get_downloads()
        options = server.api.get_options(downloads)[0]
        assert options.max_download_limit == 10000


def test_get_stats_method(server: Aria2Server) -> None:
    assert server.api.get_stats()


def test_move_method(tmp_path: Path, port: int) -> None:
    with Aria2Server(tmp_path, port, session="2-dls-paused.txt") as server:
        downloads = server.api.get_downloads()
        assert server.api.move(downloads[0], 1) == 1
        new_pos_downloads = server.api.get_downloads()
        assert downloads == list(reversed(new_pos_downloads))


def test_move_down_method(tmp_path: Path, port: int) -> None:
    with Aria2Server(tmp_path, port, session="2-dls-paused.txt") as server:
        downloads = server.api.get_downloads()
        assert server.api.move_down(downloads[0]) == 1
        new_pos_downloads = server.api.get_downloads()
        assert downloads == list(reversed(new_pos_downloads))


def test_move_to_method(tmp_path: Path, port: int) -> None:
    with Aria2Server(tmp_path, port, session="2-dls-paused.txt") as server:
        downloads = server.api.get_downloads()
        assert server.api.move_to(downloads[0], 1) == 1
        new_pos_downloads = server.api.get_downloads()
        assert new_pos_downloads == [downloads[1], downloads[0]]

        assert server.api.move_to(downloads[1], -1) == 1
        new_pos_downloads = server.api.get_downloads()
        assert new_pos_downloads == downloads


def test_move_to_bottom_method(tmp_path: Path, port: int) -> None:
    with Aria2Server(tmp_path, port, session="2-dls-paused.txt") as server:
        downloads = server.api.get_downloads()
        assert server.api.move_to_bottom(downloads[0]) == 1
        new_pos_downloads = server.api.get_downloads()
        assert new_pos_downloads == [downloads[1], downloads[0]]


def test_move_to_top_method(tmp_path: Path, port: int) -> None:
    with Aria2Server(tmp_path, port, session="2-dls-paused.txt") as server:
        downloads = server.api.get_downloads()
        assert server.api.move_to_top(downloads[1]) == 0
        new_pos_downloads = server.api.get_downloads()
        assert new_pos_downloads == [downloads[1], downloads[0]]


def test_move_up_method(tmp_path: Path, port: int) -> None:
    with Aria2Server(tmp_path, port, session="2-dls-paused.txt") as server:
        downloads = server.api.get_downloads()
        assert server.api.move_up(downloads[0]) == 0
        new_pos_downloads = server.api.get_downloads()
        assert downloads == new_pos_downloads


def test_pause_method(tmp_path: Path, port: int) -> None:
    with Aria2Server(tmp_path, port, session="big-download.txt") as server:
        time.sleep(0.1)
        downloads = server.api.get_downloads()
        if downloads[0].has_failed:
            pytest.xfail("Failed to establish connection (sporadic error)")
        assert server.api.pause([downloads[0]])
        assert downloads[0].live.status == "paused"


def test_pause_all_method(tmp_path: Path, port: int) -> None:
    with Aria2Server(tmp_path, port, session="3-dls.txt") as server:
        assert server.api.pause_all()
        # The following code block is commented out because we cannot ensure
        # that the downloads will be paused in sufficient time.
        # aria2c returns "OK" immediately, and only then proceeds to pause the downloads
        # (with potential additional steps like contacting trackers).
        # Therefore we simply check the the call goes well.

        # for _ in range(5):
        #     time.sleep(1)
        #     downloads = server.api.get_downloads()
        #     try:
        #         assert all([d.is_paused for d in downloads])
        #     except AssertionError:
        #         pass
        #     else:
        #         break
        # else:
        #     raise AssertionError


def test_autopurge_method(tmp_path: Path, port: int) -> None:
    with Aria2Server(tmp_path, port, session="3-dls.txt") as server:
        assert server.api.autopurge()


def test_remove_method(tmp_path: Path, port: int) -> None:
    with Aria2Server(tmp_path, port, session="3-dls.txt") as server:
        downloads = server.api.get_downloads()
        if not all(server.api.remove(downloads)):
            pytest.xfail("Sporadic failures")
        downloads = server.api.get_downloads()
        assert not downloads


def test_remove_files_method(tmp_path: Path, port: int) -> None:
    with Aria2Server(tmp_path, port, session="very-small-download.txt") as server:
        time.sleep(1)
        download = server.api.get_downloads()[0]
        while not download.live.is_complete:
            if download.has_failed:
                pytest.xfail("Failed to establish connection (sporadic error)")
            time.sleep(0.1)
        assert server.api.remove([download], files=True)
        for file in download.root_files_paths:
            assert not file.exists()


def test_remove_files_not_complete(tmp_path: Path, port: int) -> None:
    with Aria2Server(tmp_path, port, session="2-dls-paused.txt") as server:
        downloads = server.api.get_downloads()
        assert server.api.remove(downloads, files=True)
        for download in downloads:
            for file in download.root_files_paths:
                assert file.exists()


def test_remove_files_tree(server: Aria2Server) -> None:
    directory = server.tmp_dir / "some-directory"
    directory.mkdir()

    class _Download:
        is_complete = True
        root_files_paths = [directory]  # noqa: RUF012

    assert server.api.remove_files([_Download])  # type: ignore[list-item]
    assert not directory.exists()


def test_remove_all_method(tmp_path: Path, port: int) -> None:
    with Aria2Server(tmp_path, port, session="3-dls.txt") as server:
        if not server.api.remove_all():
            pytest.xfail("Sporadic failures")
        downloads = server.api.get_downloads()
        assert not downloads


def test_resume_method(tmp_path: Path, port: int) -> None:
    with Aria2Server(tmp_path, port, session="2-dls-paused.txt") as server:
        downloads = server.api.get_downloads()
        assert server.api.resume(downloads)
        downloads = server.api.get_downloads()
        active = [d.is_active for d in downloads]
        if not all(active):
            pytest.xfail("Not all downloads were resumed (sporadic error)")
        assert all(active)


def test_resume_all_method(tmp_path: Path, port: int) -> None:
    with Aria2Server(tmp_path, port, session="2-dls-paused.txt") as server:
        time.sleep(0.1)
        assert server.api.resume_all()
        time.sleep(0.1)
        downloads = server.api.get_downloads()
        for download in downloads:
            if download.has_failed:
                pytest.xfail("Failed to establish connection (sporadic error)")
        assert all(d.is_active for d in downloads)


def test_set_global_options_method(tmp_path: Path, port: int) -> None:
    with Aria2Server(tmp_path, port, config=CONFIGS_DIR / "max-5-dls.conf") as server:
        assert server.api.set_global_options({"max-concurrent-downloads": "10"})
        options = server.api.get_global_options()
        assert options.max_concurrent_downloads == 10


def test_set_options_method(tmp_path: Path, port: int) -> None:
    with Aria2Server(tmp_path, port, session="max-dl-limit-10000.txt") as server:
        downloads = server.api.get_downloads()
        try:
            assert server.api.set_options({"max-download-limit": "20000"}, downloads)[0]
        except ClientException:
            pytest.xfail("Cannot change option for some reason")
        try:
            options = server.api.get_options(downloads)[0]
        except ClientException:
            pytest.xfail("Cannot change option for some reason")
        assert options.max_download_limit == 20000


# @pytest.mark.flaky(reruns=5)
def test_copy_files_method(tmp_path_factory: pytest.TempPathFactory, port: int) -> None:
    with Aria2Server(tmp_path_factory.mktemp("copy_files"), port, session="very-small-download.txt") as server:
        # initialize temp dir to copy to
        tmp_dir = tmp_path_factory.mktemp("copy_files")

        # wait until download is finished
        download = server.api.get_downloads()[0]
        while not download.live.is_complete:
            if download.has_failed:
                pytest.xfail("Failed to establish connection (sporadic error)")
            time.sleep(0.2)

        # actual method run
        server.api.copy_files([download], tmp_dir)

        # assert file was copied and contents are identical
        source = download.files[0].path
        target = tmp_dir / source.name

        assert source.exists()
        assert target.exists()

        with open(source) as stream:
            source_contents = stream.read()
        with open(target) as stream:
            target_contents = stream.read()
        assert source_contents == target_contents

        # clean up
        target.unlink()
        tmp_dir.rmdir()


def test_move_files_method(tmp_path_factory: pytest.TempPathFactory, port: int) -> None:
    with Aria2Server(tmp_path_factory.mktemp("move_files"), port, session="very-small-download.txt") as server:
        # initialize temp dir to copy to
        tmp_dir = tmp_path_factory.mktemp("move_files")

        # wait until download is finished
        download = server.api.get_downloads()[0]
        while not download.live.is_complete:
            if download.has_failed:
                pytest.xfail("Failed to establish connection (sporadic error)")
            time.sleep(0.2)

        # read source contents before move
        source = download.files[0].path
        with open(source) as stream:
            source_contents = stream.read()

        # actual method run
        server.api.move_files([download], tmp_dir)

        # assert file was copied and contents are identical
        target = tmp_dir / source.name

        assert not source.exists()
        assert target.exists()

        with open(target) as stream:
            target_contents = stream.read()
        assert source_contents == target_contents

        # clean up
        target.unlink()
        tmp_dir.rmdir()


def test_listen_to_notifications(tmp_path: Path, port: int) -> None:
    with Aria2Server(tmp_path, port, session="2-dls-paused.txt") as server:
        server.api.listen_to_notifications(threaded=True, timeout=1)
    time.sleep(3)
    assert server.api.listener
    assert not server.api.listener.is_alive()


def test_listen_to_notifications_then_stop(port: int) -> None:
    api = API(Client(port=port))
    api.listen_to_notifications(threaded=True, timeout=1)
    api.stop_listening()
    assert api.listener is None


def test_listen_to_notifications_callbacks(tmp_path: Path, port: int, capsys: pytest.CaptureFixture) -> None:
    with Aria2Server(tmp_path, port, session="2-dls-paused.txt") as server:
        server.api.listen_to_notifications(
            on_download_start=lambda api, gid: print("started " + gid),  # noqa: T201
            threaded=True,
            timeout=1,
        )
        time.sleep(1)
        server.api.resume_all()
        time.sleep(3)
        server.api.stop_listening()
    assert capsys.readouterr().out == "started 0000000000000001\nstarted 0000000000000002\n"


def test_listen_to_notifications_no_thread(tmp_path: Path, port: int) -> None:
    with Aria2Server(tmp_path, port, session="2-dls-paused.txt") as server:

        def thread_target() -> None:
            server.api.listen_to_notifications(threaded=False, timeout=1)

        thread = threading.Thread(target=thread_target)
        thread.start()
        time.sleep(1)
        server.client.stop_listening()
        time.sleep(1)
        server.api.stop_listening()


def test_parse_input_file() -> None:
    api = API()

    downloads = api.parse_input_file(INPUT_FILES[0])
    assert len(downloads) == 2

    downloads = api.parse_input_file(INPUT_FILES[1])
    assert len(downloads) == 1

    downloads = api.parse_input_file(INPUT_FILES[2])
    assert len(downloads) == 0



"""The main CLI function."""

from __future__ import annotations

import sys
from pathlib import Path

import requests
from loguru import logger

from aria2p import enable_logger
from aria2p.api import API
from aria2p.cli.commands.add import add
from aria2p.cli.commands.add_magnet import add_magnets
from aria2p.cli.commands.add_metalink import add_metalinks
from aria2p.cli.commands.add_torrent import add_torrents
from aria2p.cli.commands.call import call
from aria2p.cli.commands.listen import listen
from aria2p.cli.commands.pause import pause
from aria2p.cli.commands.purge import purge
from aria2p.cli.commands.remove import remove
from aria2p.cli.commands.resume import resume
from aria2p.cli.commands.show import show
from aria2p.cli.commands.top import top
from aria2p.cli.parser import check_args, get_parser
from aria2p.client import Client, ClientException

commands = {
    None: top,  # default command
    "show": show,
    "top": top,
    "call": call,
    "add": add,
    "add-magnet": add_magnets,  # alias for add-magnets
    "add-magnets": add_magnets,
    "add-torrent": add_torrents,  # alias for add-torrents
    "add-torrents": add_torrents,
    "add-metalink": add_metalinks,  # alias for add-metalinks
    "add-metalinks": add_metalinks,
    "pause": pause,
    "stop": pause,  # alias for pause
    "resume": resume,
    "start": resume,  # alias for resume
    "remove": remove,
    "rm": remove,  # alias for remove
    "del": remove,  # alias for remove
    "delete": remove,  # alias for remove
    "purge": purge,
    "autopurge": purge,  # alias for purge
    "autoclear": purge,  # alias for purge
    "autoremove": purge,  # alias for purge
    "listen": listen,
}


def main(args: list[str] | None = None) -> int:
    """Run the main program.

    This function is executed when you type `aria2p` or `python -m aria2p`.

    Parameters:
        args: Parameters passed from the command line.

    Returns:
        An exit code.
    """
    parser = get_parser()
    opts = parser.parse_args(args=args)
    kwargs = opts.__dict__

    log_level = kwargs.pop("log_level")
    log_path = kwargs.pop("log_path")

    if log_path:
        log_path = Path(log_path)
        if log_path.is_dir():
            log_path = log_path / "aria2p-{time}.log"
        enable_logger(sink=log_path, level=log_level or "WARNING")
    elif log_level:
        enable_logger(sink=sys.stderr, level=log_level)

    logger.debug("Checking arguments")
    check_args(parser, opts)

    logger.debug("Instantiating API")
    api = API(
        Client(
            host=kwargs.pop("host"),
            port=kwargs.pop("port"),
            secret=kwargs.pop("secret"),
            timeout=kwargs.pop("client_timeout"),
        ),
    )

    logger.info(f"API instantiated: {api!r}")

    # Warn if no aria2 daemon process seems to be running
    logger.debug("Testing connection")
    try:
        api.client.get_version()
    except requests.ConnectionError as error:
        print(f"[ERROR] {error}", file=sys.stderr)
        print(file=sys.stderr)
        print("Please make sure that an instance of aria2c is running with RPC mode enabled,", file=sys.stderr)
        print("and that you have provided the right host, port and secret token.", file=sys.stderr)
        print("More information at https://pawamoy.github.io/aria2p.", file=sys.stderr)
        return 2

    subcommand = kwargs.pop("subcommand")
    kwargs.pop("debug_info")

    if subcommand:
        logger.debug("Running subcommand " + subcommand)
    try:
        return commands[subcommand](api, **kwargs)  # type: ignore
    except ClientException as error:
        print(str(error), file=sys.stderr)
        return error.code



"""Command to remove downloads."""

from __future__ import annotations

import sys
from typing import TYPE_CHECKING

from aria2p.client import ClientException

if TYPE_CHECKING:
    from aria2p.api import API


def remove(
    api: API,
    gids: list[str] | None = None,
    do_all: bool = False,  # noqa: FBT001,FBT002
    force: bool = False,  # noqa: FBT001,FBT002
) -> int:
    """Remove subcommand.

    Parameters:
        api: The API instance to use.
        gids: The GIDs of the downloads to remove.
        do_all: Pause all downloads if True.
        force: Force pause or not (see API.remove).

    Returns:
        int: 0 if all success, 1 if one failure.
    """
    if do_all:
        if api.remove_all():
            return 0
        return 1

    try:
        downloads = api.get_downloads(gids)
    except ClientException as error:
        print(str(error), file=sys.stderr)
        return 1

    ok = True
    result = api.remove(downloads, force=force)

    if all(result):
        return 0 if ok else 1

    for item in result:
        if isinstance(item, ClientException):
            print(item, file=sys.stderr)

    return 1



"""Command to call RPC methods."""

from __future__ import annotations

import json
import sys
from typing import TYPE_CHECKING, Any

from aria2p.client import Client

if TYPE_CHECKING:
    from aria2p.api import API


def get_method(name: str) -> str | None:
    """Return the actual aria2 method name from a differently formatted name.

    Parameters:
        name: A method name.

    Returns:
        The real method name.
    """
    methods = {}

    for method in Client.METHODS:
        methods[method.lower()] = method
        methods[method.split(".")[1].lower()] = method

    name = name.lower()
    name = name.replace("-", "")
    name = name.replace("_", "")

    return methods.get(name)


def call(api: API, method: str, params: str | list[str]) -> int:
    """Call subcommand.

    Parameters:
        api: The API instance to use.
        method: Name of the method to call.
        params: Parameters to use when calling method.

    Returns:
        int: Always 0.
    """
    real_method = get_method(method)

    if real_method is None:
        print(f"aria2p: call: Unknown method {method}.", file=sys.stderr)
        print("  Run 'aria2p call listmethods' to list the available methods.", file=sys.stderr)
        return 1

    call_params: list[Any]
    if isinstance(params, str):
        call_params = json.loads(params)
    elif params is None:
        call_params = []
    else:
        call_params = params

    response = api.client.call(real_method, call_params)
    print(json.dumps(response))

    return 0



"""Tests for the `client` module."""

from __future__ import annotations

import json
import os
import signal
import threading
import time
from base64 import b64encode
from copy import deepcopy
from typing import TYPE_CHECKING

import pytest
import requests
from responses import mock as responses

from aria2p import Client, ClientException
from aria2p.client import JSONRPC_CODES, JSONRPC_PARSER_ERROR, Notification
from tests import BUNSENLABS_MAGNET, BUNSENLABS_TORRENT, CONFIGS_DIR, DEBIAN_METALINK, SESSIONS_DIR, XUBUNTU_MIRRORS
from tests.conftest import Aria2Server

if TYPE_CHECKING:
    from pathlib import Path


class TestParameters:
    # callback that return params of a single call as result
    @staticmethod
    def call_params_callback(request: requests.PreparedRequest) -> tuple[int, dict, str]:
        payload = json.loads(request.body)  # type: ignore[arg-type]
        resp_body = {"result": payload["params"]}
        return 200, {}, json.dumps(resp_body)

    # callback that return params of a batch call as result
    @staticmethod
    def batch_call_params_callback(request: requests.PreparedRequest) -> tuple[int, dict, str]:
        payload = json.loads(request.body)  # type: ignore[arg-type]
        resp_body = [{"result": method["params"]} for method in payload]
        return 200, {}, json.dumps(resp_body)

    @responses.activate
    def test_insert_secret_with_aria2_method_call(self) -> None:
        # create client with secret
        secret = "hello"  # noqa: S105
        client = Client(secret=secret)

        responses.add_callback(responses.POST, client.server, callback=self.call_params_callback)

        # create params
        params = ["param1", "param2"]
        # copy params and insert secret
        expected_params = deepcopy(params)
        expected_params.insert(0, f"token:{secret}")

        # call function and assert result
        resp = client.call(client.ADD_URI, params, insert_secret=True)
        assert resp == expected_params

    @responses.activate
    def test_insert_secret_with_system_multicall(self) -> None:
        # create client with secret
        secret = "hello"  # noqa: S105
        client = Client(secret=secret)

        responses.add_callback(responses.POST, client.server, callback=self.call_params_callback)

        # create params
        params = [
            [
                {"methodName": client.ADD_URI, "params": ["param1", "param2"]},
                {"methodName": client.ADD_URI, "params": ["param3", "param4"]},
            ],
        ]
        # copy params and insert secret
        expected_params = deepcopy(params)
        for param in expected_params[0]:
            param["params"].insert(0, f"token:{secret}")  # type: ignore[attr-defined]

        # call function and assert result
        resp = client.call(client.MULTICALL, params, insert_secret=True)
        assert resp == expected_params

    @responses.activate
    def test_does_not_insert_secret_with_unknown_method_call(self) -> None:
        # create client with secret
        secret = "hello"  # noqa: S105
        client = Client(secret=secret)

        responses.add_callback(responses.POST, client.server, callback=self.call_params_callback)

        # create params
        params = ["param1", "param2"]

        # call function and assert result
        resp = client.call("other.method", params, insert_secret=True)
        assert secret not in resp  # type: ignore[operator]

    @responses.activate
    def test_does_not_insert_secret_if_told_so(self) -> None:
        # create client with secret
        secret = "hello"  # noqa: S105
        client = Client(secret=secret)

        responses.add_callback(responses.POST, client.server, callback=self.call_params_callback)

        # create params
        params = ["param1", "param2"]

        # call function and assert result
        resp = client.call("other.method", params, insert_secret=False)
        assert secret not in resp  # type: ignore[operator]

    def test_client_str_returns_client_server(self) -> None:
        host = "https://localhost:8779/"
        port = 7100
        client = Client(host, port)
        assert client.server == f"{host.rstrip('/')}:{port}/jsonrpc" == str(client)

    @responses.activate
    def test_batch_call(self) -> None:
        client = Client()

        responses.add_callback(responses.POST, client.server, callback=self.batch_call_params_callback)

        # create params
        params_1 = ["param1", "param2"]
        params_2 = ["param3", "param4"]
        # copy params and insert secret
        expected_params = [params_1, params_2]

        # call function and assert result
        resp = client.batch_call([(client.ADD_URI, params_1, 0), (client.ADD_METALINK, params_2, 1)])
        assert resp == expected_params

    @responses.activate
    def test_insert_secret_with_batch_call(self) -> None:
        # create client with secret
        secret = "hello"  # noqa: S105
        client = Client(secret=secret)

        responses.add_callback(responses.POST, client.server, callback=self.batch_call_params_callback)

        # create params
        params_1 = ["param1", "param2"]
        params_2 = ["param3", "param4"]
        # copy params and insert secret
        expected_params = [deepcopy(params_1), deepcopy(params_2)]
        for p in expected_params:
            p.insert(0, f"token:{secret}")

        # call function and assert result
        resp = client.batch_call(
            [(client.ADD_URI, params_1, 0), (client.ADD_METALINK, params_2, 1)],
            insert_secret=True,
        )
        assert resp == expected_params

    @responses.activate
    def test_multicall2(self) -> None:
        client = Client()

        responses.add_callback(responses.POST, client.server, callback=self.call_params_callback)

        # create params
        params_1 = ["0000000000000001"]
        params_2 = ["2fa07b6e85c40205"]
        calls = [(client.REMOVE, params_1), (client.REMOVE, params_2)]
        # copy params and insert secret
        expected_params = [
            [
                {"methodName": client.REMOVE, "params": deepcopy(params_1)},
                {"methodName": client.REMOVE, "params": deepcopy(params_2)},
            ],
        ]

        # call function and assert result
        resp = client.multicall2(calls)
        assert resp == expected_params

    @responses.activate
    def test_insert_secret_with_multicall2(self) -> None:
        # create client with secret
        secret = "hello"  # noqa: S105
        client = Client(secret=secret)

        responses.add_callback(responses.POST, client.server, callback=self.call_params_callback)

        # create params
        params_1 = ["0000000000000001"]
        params_2 = ["2fa07b6e85c40205"]
        calls = [(client.REMOVE, params_1), (client.REMOVE, params_2)]
        # copy params and insert secret
        expected_params = [
            [
                {"methodName": client.REMOVE, "params": deepcopy(params_1)},
                {"methodName": client.REMOVE, "params": deepcopy(params_2)},
            ],
        ]
        for param in expected_params[0]:
            param["params"].insert(0, f"token:{secret}")  # type: ignore[attr-defined]

        # call function and assert result
        resp = client.multicall2(calls, insert_secret=True)
        assert resp == expected_params


class TestClientExceptionClass:
    @responses.activate
    def test_call_raises_custom_error(self) -> None:
        client = Client()
        responses.add(
            responses.POST,
            client.server,
            json={"error": {"code": 1, "message": "Custom message"}},
            status=200,
        )
        with pytest.raises(ClientException, match=r"Custom message") as e:  # noqa: PT012
            client.call("aria2.method")
            assert e.code == 1  # type: ignore[attr-defined]

    @responses.activate
    def test_call_raises_known_error(self) -> None:
        client = Client()
        responses.add(
            responses.POST,
            client.server,
            json={"error": {"code": JSONRPC_PARSER_ERROR, "message": "Custom message"}},
            status=200,
        )
        with pytest.raises(  # noqa: PT012
            ClientException,
            match=rf"{JSONRPC_CODES[JSONRPC_PARSER_ERROR]}\nCustom message",
        ) as e:
            client.call("aria2.method")
            assert e.code == JSONRPC_PARSER_ERROR  # type: ignore[attr-defined]


class TestClientClass:
    def test_add_metalink_method(self, server: Aria2Server) -> None:
        # get file contents
        with open(DEBIAN_METALINK, "rb") as stream:
            metalink_contents = stream.read()
        encoded_contents = b64encode(metalink_contents).decode("utf-8")

        assert server.client.add_metalink(encoded_contents)

    def test_add_torrent_method(self, server: Aria2Server) -> None:
        # get file contents
        with open(BUNSENLABS_TORRENT, "rb") as stream:
            torrent_contents = stream.read()
        encoded_contents = b64encode(torrent_contents).decode("utf-8")

        assert server.client.add_torrent(encoded_contents, [])

    def test_add_uri_method(self, server: Aria2Server) -> None:
        assert server.client.add_uri([BUNSENLABS_MAGNET])
        assert server.client.add_uri(XUBUNTU_MIRRORS)

    def test_global_option_methods(self, tmp_path: Path, port: int) -> None:
        with Aria2Server(tmp_path, port, config=CONFIGS_DIR / "max-5-dls.conf") as server:
            max_concurrent_downloads = server.client.get_global_option()["max-concurrent-downloads"]
            assert max_concurrent_downloads == "5"

            assert server.client.change_global_option({"max-concurrent-downloads": "10"}) == "OK"

            max_concurrent_downloads = server.client.get_global_option()["max-concurrent-downloads"]
            assert max_concurrent_downloads == "10"

    def test_option_methods(self, tmp_path: Path, port: int) -> None:
        with Aria2Server(tmp_path, port, session="max-dl-limit-10000.txt") as server:
            time.sleep(0.1)
            try:
                gid = server.client.tell_active(keys=["gid"])[0]["gid"]
            except IndexError:
                pytest.xfail("Failed to establish connection (sporadic error)")
            max_download_limit = server.client.get_option(gid=gid)["max-download-limit"]
            assert max_download_limit == "10000"

            assert server.client.change_option(gid, {"max-download-limit": "20000"}) == "OK"

            max_download_limit = server.client.get_option(gid)["max-download-limit"]
            assert max_download_limit == "20000"

    def test_position_method(self, tmp_path: Path, port: int) -> None:
        with Aria2Server(tmp_path, port, session="2-dls-paused.txt") as server:
            gids = server.client.tell_waiting(0, 5, keys=["gid"])
            first, second = (r["gid"] for r in gids)
            assert server.client.change_position(second, 0, "POS_SET") == 0
            assert server.client.change_position(second, 5, "POS_CUR") == 1

    def test_change_uri_method(self, tmp_path: Path, port: int) -> None:
        with Aria2Server(tmp_path, port, session="1-dl-2-uris.txt") as server:
            gid = server.client.tell_waiting(0, 1, keys=["gid"])[0]["gid"]
            assert server.client.change_uri(gid, 1, ["http://localhost:8779/1024"], ["http://localhost:8779/1k"]) == [
                1,
                1,
            ]
            assert server.client.change_uri(gid, 1, ["http://localhost:8779/1k"], []) == [1, 0]

    def test_force_pause_method(self, tmp_path: Path, port: int) -> None:
        with Aria2Server(tmp_path, port, session="big-download.txt") as server:
            time.sleep(0.1)
            try:
                gid = server.client.tell_active(keys=["gid"])[0]["gid"]
            except IndexError:
                pytest.xfail("Failed to establish connection (sporadic error)")
            assert server.client.force_pause(gid) == gid

    def test_force_pause_all_method(self, tmp_path: Path, port: int) -> None:
        with Aria2Server(tmp_path, port, session="2-dls.txt") as server:
            assert server.client.force_pause_all() == "OK"

    def test_force_remove_method(self, tmp_path: Path, port: int) -> None:
        with Aria2Server(tmp_path, port, session="big-download.txt") as server:
            try:
                gid = server.client.tell_active(keys=["gid"])[0]["gid"]
            except IndexError:
                pytest.xfail("Failed to establish connection (sporadic error)")
            assert server.client.force_remove(gid)
            assert server.client.tell_status(gid, keys=["status"])["status"] == "removed"

    def test_force_shutdown_method(self, server: Aria2Server) -> None:
        assert server.client.force_shutdown() == "OK"
        with pytest.raises(requests.ConnectionError):  # noqa: PT012
            for _retry in range(10):
                server.client.list_methods()
                time.sleep(1)

    def test_get_files_method(self, tmp_path: Path, port: int) -> None:
        with Aria2Server(tmp_path, port, session="1-dl.txt") as server:
            active = server.client.tell_active(keys=["gid"])
            if not active:
                pytest.xfail("Index error (sporadic error)")
            assert len(server.client.get_files(active[0]["gid"])) == 1

    def test_get_global_stat_method(self, server: Aria2Server) -> None:
        assert server.client.get_global_stat()

    def test_get_peers_method(self, tmp_path: Path, port: int) -> None:
        with Aria2Server(tmp_path, port, session="max-dl-limit-10000.txt") as server:
            time.sleep(0.1)
            try:
                gid = server.client.tell_active(keys=["gid"])[0]["gid"]
            except IndexError:
                pytest.xfail("Failed to establish connection (sporadic error)")
            assert not server.client.get_peers(gid)

    def test_get_servers_method(self, tmp_path: Path, port: int) -> None:
        with Aria2Server(tmp_path, port, session="max-dl-limit-10000.txt") as server:
            time.sleep(0.1)
            try:
                gid = server.client.tell_active(keys=["gid"])[0]["gid"]
            except IndexError:
                pytest.xfail("Failed to establish connection (sporadic error)")
            assert server.client.get_servers(gid)

    def test_get_session_info_method(self, server: Aria2Server) -> None:
        assert server.client.get_session_info()

    def test_get_uris_method(self, tmp_path: Path, port: int) -> None:
        with Aria2Server(tmp_path, port, session="1-dl-2-uris.txt") as server:
            gid = server.client.tell_waiting(0, 1, keys=["gid"])[0]["gid"]
            assert server.client.get_uris(gid) == [
                {"status": "waiting", "uri": "http://localhost:8779/1024"},
                {"status": "waiting", "uri": "http://localhost:8779/1k"},
            ]

    def test_get_version_method(self, server: Aria2Server) -> None:
        assert server.client.get_version()

    def test_list_methods_method(self, server: Aria2Server) -> None:
        assert server.client.list_methods()

    def test_list_notifications_method(self, server: Aria2Server) -> None:
        assert server.client.list_notifications()

    def test_multicall_method(self, server: Aria2Server) -> None:
        assert server.client.multicall(
            [{"methodName": server.client.LIST_METHODS}, {"methodName": server.client.LIST_NOTIFICATIONS}],
        )

    def test_multicall2_method(self, server: Aria2Server) -> None:
        assert server.client.multicall2([(server.client.LIST_METHODS, []), (server.client.LIST_NOTIFICATIONS, [])])

    def test_pause_method(self, tmp_path: Path, port: int) -> None:
        with Aria2Server(tmp_path, port, session="1-dl.txt") as server:
            time.sleep(0.1)
            try:
                gid = server.client.tell_active(keys=["gid"])[0]["gid"]
            except IndexError:
                pytest.xfail("Failed to establish connection (sporadic error)")
            assert server.client.pause(gid) == gid

    def test_pause_all_method(self, tmp_path: Path, port: int) -> None:
        with Aria2Server(tmp_path, port, session="2-dls.txt") as server:
            assert server.client.pause_all() == "OK"

    def test_purge_download_result_method(self, server: Aria2Server) -> None:
        assert server.client.purge_download_result() == "OK"

    def test_remove_method(self, tmp_path: Path, port: int) -> None:
        with Aria2Server(tmp_path, port, session="1-dl.txt") as server:
            time.sleep(0.1)
            try:
                gid = server.client.tell_active(keys=["gid"])[0]["gid"]
            except IndexError:
                pytest.xfail("Failed to establish connection (sporadic error)")
            assert server.client.remove(gid)
            assert server.client.tell_status(gid, keys=["status"])["status"] == "removed"

    def test_remove_download_result_method(self, tmp_path: Path, port: int) -> None:
        with Aria2Server(tmp_path, port, session="1-dl.txt") as server:
            time.sleep(0.1)
            try:
                gid = server.client.tell_active(keys=["gid"])[0]["gid"]
            except IndexError:
                pytest.xfail("Failed to establish connection (sporadic error)")
            server.client.remove(gid)
            assert server.client.remove_download_result(gid) == "OK"
            assert len(server.client.tell_stopped(0, 1)) == 0

    def test_save_session_method(self, tmp_path: Path, port: int) -> None:
        session_input = SESSIONS_DIR / "1-dl.txt"
        with Aria2Server(tmp_path, port, session=session_input) as server:
            session_output = server.tmp_dir / "_session.txt"
            server.client.change_global_option({"save-session": str(session_output)})
            assert server.client.save_session() == "OK"
            with open(session_input) as stream:
                input_contents = stream.read()
            with open(session_output) as stream:
                output_contents = stream.read()
            for line in input_contents.split("\n"):
                assert line in output_contents

    def test_shutdown_method(self, server: Aria2Server) -> None:
        assert server.client.shutdown() == "OK"
        with pytest.raises(requests.ConnectionError):  # noqa: PT012
            for _retry in range(10):
                server.client.list_methods()
                time.sleep(1)

    def test_tell_active_method(self, tmp_path: Path, port: int) -> None:
        with Aria2Server(tmp_path, port, session="big-download.txt") as server:
            time.sleep(0.1)
            if server.api.get_download("0000000000000001").has_failed:
                pytest.xfail("Failed to establish connection (sporadic error)")
            assert len(server.client.tell_active(keys=["gid"])) > 0

    def test_tell_status_method(self, tmp_path: Path, port: int) -> None:
        with Aria2Server(tmp_path, port, session="1-dl-paused.txt") as server:
            gid = server.client.tell_waiting(0, 1, keys=["gid"])[0]["gid"]
            assert server.client.tell_status(gid)

    def test_tell_stopped_method(self, tmp_path: Path, port: int) -> None:
        with Aria2Server(tmp_path, port, session="very-small-download.txt") as server:
            download = server.api.get_download("0000000000000001")
            while not download.live.is_complete:
                if download.has_failed:
                    pytest.xfail("Failed to establish connection (sporadic error)")
                time.sleep(0.1)
            assert len(server.client.tell_stopped(0, 1, keys=["gid"])) > 0

    def test_tell_waiting_method(self, tmp_path: Path, port: int) -> None:
        with Aria2Server(tmp_path, port, session="2-dls-paused.txt") as server:
            assert server.client.tell_waiting(0, 5, keys=["gid"]) == [
                {"gid": "0000000000000001"},
                {"gid": "0000000000000002"},
            ]

    def test_unpause_method(self, tmp_path: Path, port: int) -> None:
        with Aria2Server(tmp_path, port, session="1-dl-paused.txt") as server:
            gid = server.client.tell_waiting(0, 1, keys=["gid"])[0]["gid"]
            assert server.client.unpause(gid) == gid

    def test_unpause_all_method(self, tmp_path: Path, port: int) -> None:
        with Aria2Server(tmp_path, port, session="2-dls-paused.txt") as server:
            assert server.client.unpause_all() == "OK"

    def test_listen_to_notifications_no_server(self) -> None:
        client = Client(port=7035)
        client.listen_to_notifications(timeout=1)

    def test_listen_to_notifications_no_callbacks(self, tmp_path: Path, port: int) -> None:
        with Aria2Server(tmp_path, port, session="2-dls-paused.txt") as server:

            def thread_target() -> None:
                server.client.listen_to_notifications(timeout=1, handle_signals=False)

            thread = threading.Thread(target=thread_target)
            thread.start()
            server.client.unpause("0000000000000001")
            time.sleep(3)
        thread.join()

    def test_listen_to_notifications_callbacks(self, tmp_path: Path, port: int, capsys: pytest.CaptureFixture) -> None:
        with Aria2Server(tmp_path, port, session="2-dls-paused.txt") as server:

            def thread_target() -> None:
                server.client.listen_to_notifications(
                    on_download_start=lambda gid: print("started " + gid),  # noqa: T201
                    timeout=1,
                    handle_signals=False,
                )

            thread = threading.Thread(target=thread_target)
            thread.start()
            time.sleep(1)
            server.client.unpause("0000000000000001")
            time.sleep(3)
        thread.join()
        assert capsys.readouterr().out == "started 0000000000000001\n"

    def test_listen_to_notifications_then_stop(self, tmp_path: Path, port: int) -> None:
        with Aria2Server(tmp_path, port, session="2-dls-paused.txt") as server:

            def thread_target() -> None:
                server.client.listen_to_notifications(timeout=1, handle_signals=False)

            thread = threading.Thread(target=thread_target)
            thread.start()
            server.client.stop_listening()
            thread.join()

    def test_listen_to_notifications_then_stop_with_signal(self, tmp_path: Path, port: int) -> None:
        with Aria2Server(tmp_path, port, session="2-dls-paused.txt") as server:

            def thread_target() -> None:
                time.sleep(2)
                os.kill(os.getpid(), signal.SIGTERM)

            thread = threading.Thread(target=thread_target)
            thread.start()
            server.client.listen_to_notifications(timeout=1, handle_signals=True)
            thread.join()


class TestNotificationClass:
    def test_init(self) -> None:
        notification = Notification("random", "random")
        assert notification

    def test_get(self) -> None:
        message = {"method": "random_event", "params": [{"gid": "random_gid"}]}
        assert Notification.get_or_raise(message)

    def test_raise(self) -> None:
        message = {"error": {"code": 9000, "message": "it's over 9000"}}
        with pytest.raises(ClientException):
            Notification.get_or_raise(message)


class TestSecretToken:
    def test_works_correctly_with_secret_set(self, tmp_path: Path, port: int) -> None:
        with Aria2Server(tmp_path, port, secret="this secret token") as server:  # noqa: S106
            assert server.client.get_version()

    def test_does_not_authorize_with_invalid_secret(self, tmp_path: Path, port: int) -> None:
        with Aria2Server(tmp_path, port, secret="this secret token") as server:  # noqa: S106
            server.client.secret = "invalid secret token"  # noqa: S105
            with pytest.raises(ClientException):
                server.client.get_version()
